"""
CBMPy: CBNetDB module
=====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2022 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBNetDB.py 710 2020-04-27 14:22:34Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals

import os
import time
import re
import webbrowser
import csv

try:
    from urllib2 import quote as urlquote
    from urllib2 import unquote as urlunquote
except ImportError:
    from urllib.parse import quote as urlquote
    from urllib.parse import unquote as urlunquote

HAVE_SQLITE2 = False
HAVE_SQLITE3 = False

try:
    import sqlite3 as sqlite3

    HAVE_SQLITE3 = True
except ImportError:
    HAVE_SQLITE3 = False
    try:
        from pysqlite2 import dbapi2 as sqlite2

        HAVE_SQLITE2 = True
    except ImportError:
        HAVE_SQLITE2 = False

from .CBConfig import __CBCONFIG__ as __CBCONFIG__

__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']


class NetDBbase(object):
    text_encoding = 'utf8'

    def URLEncode(self, txt):
        """
        URL encodes a string.

        """
        return urlquote(txt.encode(self.text_encoding))

    def URLDecode(self, txt):
        """
        Decodes a URL encoded string

        """
        return urlunquote(txt)


class DBTools(NetDBbase):
    """
    Tools to work with SQLite DB's (optimized, no SQL required).

    """

    sqlite = None
    sqlite_version = None
    db_conn = None
    db_cursor = None
    db_tables = None

    def __init__(self):
        if HAVE_SQLITE3:
            self.sqlite = sqlite3
            self.sqlite_version = 3
        elif HAVE_SQLITE2:
            self.sqlite = sqlite2
            self.sqlite_version = 2
        else:
            raise RuntimeError("\nSQLite not installed")

        self.db_tables = []

    def connectSQLiteDB(self, db_name, work_dir=None):
        """
        Connect to a sqlite database.

        - *db_name* the name of the sqlite database
        - *work_dir* the optional database path

        """
        # connect to DB
        if work_dir != None:
            self.db_conn = self.sqlite.connect(os.path.join(work_dir, db_name))
        else:
            self.db_conn = self.sqlite.connect(db_name)

        self.db_cursor = self.db_conn.cursor()

    def commitDB(self):
        """
        Commits all curent changes to DB, returns a boolean.
        """
        try:
            self.db_cursor.connection.commit()
            return True
        except:
            return False

    def createDBTable(self, table, sqlcols):
        """
        Create a database table if it does not exist:

         - *table* the table name
         - *sqlcols* a list containing the SQL definitions of the table columns: <id> <type> for example `['gene TEXT PRIMARY KEY', 'aa_seq TEXT', 'nuc_seq TEXT', 'aa_len INT', 'nuc_len INT']`

        Effectively writes CREATE TABLE "table" (<id> <type>, gene TEXT PRIMARY KEY, aa_seq TEXT, nuc_seq TEXT, aa_len INT, nuc_len INT) % table
        """
        SQL = 'CREATE TABLE %s (' % table
        for c in sqlcols:
            SQL += ' %s,' % c
        SQL = SQL[:-1]
        SQL += ' )'
        print(SQL)
        try:
            self.db_cursor.execute('SELECT * FROM %s' % table)
            print('Table {} exists'.format(table))
        except self.sqlite.OperationalError:
            print('Table {} does not exist, creating it'.format(table))
            self.db_cursor.execute(SQL)
        self.db_tables.append(table)

    def insertData(self, table, data, commit=True):
        """
        Insert data into a table: "INSERT INTO %s (?, ?, ?, ?, ?) VALUES (?, ?, ?, ?, ?)" % table,
                                    (?, ?, ?, ?, ?)) )

         - *table* the DB table name
         - *data* a dictionary of {id:value} pairs
         - *commit* whether to commit the data insertions

        """
        colstr = "("
        valstr = "VALUES ("
        vals = []
        # I want to use dictionaries but this just keeps things backwards compatible for a while
        try:
            data.keys()
            for d in data:
                colstr += '%s, ' % d
                valstr += '?, '
                vals.append(data[d])
        except AttributeError:
            print(
                '\n\nWARNING: data now uses a dictionary as input please update your code - see docstring for details\n'
            )
            time.sleep(3)
            for d in data:
                colstr += '%s, ' % d[0]
                valstr += '?, '
                vals.append(d[1])
        colstr = colstr[:-2] + ')'
        valstr = valstr[:-2] + ')'
        sql = "INSERT INTO %s %s %s" % (table, colstr, valstr)
        # print sql
        # time.sleep(1)
        try:
            self.db_cursor.execute(sql, vals)
            if commit:
                self.db_cursor.connection.commit()
            return True
        except AttributeError:
            return False

    def updateData(self, table, col, rid, data, commit=True):
        """
        Update already defined data

         - *table* the table name
         - *col* the column name
         - *rid* the row id to update
         - *data* a dictionary of {id:value} pairs
         - *commit* whether to commit the data updates

         UPDATE COMPANY SET ADDRESS = 'Texas' WHERE ID = 6;

        """

        sql = 'UPDATE {} SET '.format(table)
        for d in data:
            sql += '{}=\"{}\", '.format(d, str(data[d]).replace('\"', '\''))
        sql = sql[:-2] + ' WHERE {}=\"{}\"'.format(col, rid)

        # print(sql)
        try:
            self.db_cursor.execute(sql)
            if commit:
                self.db_cursor.connection.commit()
            return True
        except AttributeError:
            return False

    def checkEntryInColumn(self, table, col, rid):
        """
        Check if an entry exists in a table

        - *table* the table name
        - *col* the column name
        - *rid* the row to search for

        """
        self.db_cursor.execute(
            "SELECT count(*) FROM {} WHERE {}=\"{}\"".format(table, col, rid)
        )
        data = self.db_cursor.fetchone()[0]
        if data == 0:
            # print('There is no component named {}'.format(rid))
            return False
        else:
            # print('Component {} found in {} row(s)'.format(rid, data))
            return True

    def executeSQL(self, sql):
        """
        Execute a SQL command:

         - *sql* a string containing a SQL command

        """
        try:
            self.db_cursor.execute(sql)
            return True
        except AttributeError as ex:
            print('Error executing command')
            print(ex)
            return False

    def getColumns(self, table, cols):
        """
        Fetch the contents of one or more columns of data in a table

         - *table* the database table
         - *cols* a list of one or more column id's

        """
        sql = "SELECT "
        for c in cols:
            sql += '{}, '.format(str(c))
        sql = sql[:-2]
        sql += ' FROM {}'.format(table)
        dout = [[] for i in range(len(cols))]
        # print(sql)
        # print(dout)
        try:
            data = self.db_cursor.execute(sql).fetchall()
            for r_ in data:
                for c_ in range(len(r_)):
                    dout[c_].append(str(r_[c_]))
            del data
        except AttributeError:
            return None
        else:
            return dout

    def getRow(self, table, col, rid):
        """
        Get the table row(s) which correspond to rid in column. Returns the row(s) as a list, if the column is the primary key
        this is always a single entry.

         - *table* the database table
         - *col* the column id
         - *rid* the row index id

        """
        sql = "SELECT * FROM {} WHERE {}=\"{}\"".format(table, col, rid)
        # print(sql)
        dout = []
        try:
            data = self.db_cursor.execute(sql).fetchall()
            if len(data) > 1:
                print(
                    'INFO: getRow is returning multiple rows for query id: {}'.format(
                        rid
                    )
                )
            for r_ in data:
                dout.append([str(d) for d in r_])
            del data
        except AttributeError:
            return None
        else:
            return dout

    def getCell(self, table, col, rid, cell):
        """
        Get the table cell which correspond to rid in column. Returns the value or None

         - *table* the database table
         - *col* the column id
         - *rid* the row index id
         - *cell* the column of the cell you want tp extract

        """
        sql = "SELECT {} FROM {} WHERE {}=\"{}\"".format(cell, table, col, rid)
        # print(sql)
        data = None
        try:
            data = self.db_cursor.execute(sql).fetchone()
            if data is None:
                return None
            data = str(self.db_cursor.execute(sql).fetchone()[0])
        except AttributeError:
            return None
        return data

    def getTable(self, table, colOut=False):
        """
        Returns an entire database table

         - *table* the table name
         - *colOut* optionally return a tuple of (data,ColNames)

        """
        sql = 'SELECT * FROM %s' % table
        sql2 = "PRAGMA table_info( %s )" % table

        r = None
        col = None
        try:
            r = self.db_cursor.execute(sql).fetchall()
            if colOut:
                col = self.db_cursor.execute(sql2).fetchall()
                col = [str(a[1]) for a in col]
        except (AttributeError, sqlite3.OperationalError):
            return None
        if colOut:
            return r, col
        else:
            return r

    def dumpTableToTxt(self, table, filename):
        """
        Save a table as tab separated txt file

         - *table* the table to export
         - *filename* the filename of the table dump

        """
        data, head = self.getTable(table, colOut=True)
        data.insert(0, head)
        from .CBTools import exportLabelledLinkedList

        exportLabelledLinkedList(data, fname=filename, names=None, sep='\t')

    def dumpTableToCSV(self, table, filename):
        """
        Save a table as tab separated txt file

         - *table* the table to export
         - *filename* the filename of the table dump

        """
        data, head = self.getTable(table, colOut=True)
        data.insert(0, head)

        F = open(filename, 'w')
        csvw = csv.writer(F, dialect='excel')
        csvw.writerows(data)
        F.close()
        del csvw

    def fetchAll(self, sql):
        """Raw SQL query e.g. 'SELECT id FROM gene WHERE gene=\"G\"' """
        # print(sql)
        r = None
        try:
            r = self.db_cursor.execute(sql).fetchall()
        except AttributeError as ex:
            print(ex)
        return r

    def closeDB(self):
        """Close the DB connection and reset the DBTools instance (can be reconnected)"""
        self.db_conn.close()
        self.db_conn = None
        self.db_cursor = None
        self.db_tables = []


class KeGGTools(object):
    """
    Class that holds useful methods for querying KeGG via a SUDS provided soap client
    """

    Kclient = None

    def __init__(self, url):
        import suds

        self.Kclient = suds.client.Client(url)

    def fetchSeqfromKeGG(self, k_gene):
        """
        Given a gene name try and retrieve the gene and amino acid sequence
        """
        g2 = 'None'
        p2 = 'None'
        try:
            g = self.Kclient.service.bget("-f -n n %s" % k_gene)
            if g == None:
                print(
                    '\n*****\nWARNING: potential naming error in gene: {}!!\n*****\n'.format(
                        k_gene
                    )
                )
            g2 = g.split('(N)')[1].replace('\n', '')
        except AttributeError as ex:
            print('\nGene sequence get exception ({})!\n'.format(k_gene))
            print(ex)
        try:
            p = self.Kclient.service.bget("-f -n a %s" % k_gene)
            if p == None:
                print(
                    '\n*****\nWARNING: potential naming error in gene: {}!!\n*****\n'.format(
                        k_gene
                    )
                )
            p2 = p.split('(A)')[1].replace('\n', '')
        except AttributeError as ex:
            print('\nProtein sequence get exception ({})!\n'.format(k_gene))
            print(ex)
        return g2, p2


class KeGGSequenceTools(object):
    """
    Using the KeGG connector this class provides tools to construct an organims specific sequence database
    """

    DB = None
    KEGG = None

    def __init__(self, url, db_name, work_dir):
        self.DB = DBTools()
        self.DB.connectSQLiteDB(db_name, work_dir)
        self.KEGG = KeGGTools(url)

    def buildGeneDatabase(
        self, genes, tablename, UPDATE_IF_EXISTS=False, default_length=0
    ):
        cntr = 1
        cntr2 = 1

        for ecg in genes:
            print('Processing gene {} of {}'.format(cntr, len(genes)))
            entry_exists = False
            testg = self.DB.db_cursor.execute(
                'SELECT * FROM %s WHERE gene="%s" ' % (tablename, ecg)
            ).fetchall()
            if len(testg) > 0:
                entry_exists = True

            # if ecg in ['eco:b1898','eco:b1899','eco:b3692','eco:b3111','eco:b4228','eco:b2978',\
            # 'eco:b1416','eco:b3112','eco:b1417','eco:b3768','eco:b3767','eco:b4229']:
            # raw_input(testg)

            tstart = time.time()
            if not entry_exists:
                print('\tadding gene {}'.format(ecg), end=" ")
                gene2, prot2 = self.KEGG.fetchSeqfromKeGG(ecg)
                if gene2 != 'None' and prot2 != 'None':
                    self.DB.db_cursor.execute(
                        "INSERT INTO %s (gene, aa_seq, nuc_seq, aa_len, nuc_len) VALUES (?, ?, ?, ?, ?)"
                        % tablename,
                        (
                            str(ecg),
                            str(prot2),
                            str(gene2),
                            int(len(prot2)),
                            int(len(gene2)),
                        ),
                    )
                else:
                    print(
                        '\nGene {} cannot be found and is probably an incorrect annotation assigning length: {}\n'.format(
                            ecg, default_length
                        )
                    )
                    self.DB.db_cursor.execute(
                        "INSERT INTO %s (gene, aa_seq, nuc_seq, aa_len, nuc_len) VALUES (?, ?, ?, ?, ?)"
                        % tablename,
                        (str(ecg), 'None', 'None', default_length, default_length),
                    )
            elif entry_exists and UPDATE_IF_EXISTS:
                print('\tupdating gene {}'.format(ecg), end=" ")
                gene2, prot2 = self.KEGG.fetchSeqfromKeGG(ecg)
                self.DB.db_cursor.execute(
                    'UPDATE %s SET aa_seq="%s", nuc_seq="%s", aa_len="%s", nuc_len="%s" WHERE gene="%s"'
                    % (tablename, prot2, gene2, int(len(prot2)), int(len(gene2)), ecg)
                )
            else:
                print('\tskipping gene {}'.format(ecg), end=" ")
            tend = time.time()
            print(' ... done ({}).'.format(tend - tstart))
            #  if cntr == 6:
            #  break
            cntr += 1
            cntr2 += 1
            if cntr2 == 21:
                self.DB.db_cursor.connection.commit()
                cntr2 = 1

    def getPeptideLengthsFromDB(self, genes, keg_prefix):
        gene_peplen = {}
        for G in genes:
            print(G)
            ##  Glen = self.cursor.execute('SELECT aa_len FROM gene_data WHERE gene="%s"' % G).fetchall()[0][0]
            Glen = self.DB.fetchAll('SELECT aa_len FROM gene_data WHERE gene="%s"' % G)[
                0
            ][0]
            print(Glen)
            gene_peplen.update({G.replace(keg_prefix, ''): Glen})
        return gene_peplen


class RESTClient(NetDBbase):
    """
    Class that provides the basis for application specific connectors to REST web services
    """

    site_root = None
    conn = None
    history = ''
    CONNECTED = False
    USER_AGENT = (
        'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:13.0) Gecko/20100101 Firefox/13.0'
    )

    def __init__(self):
        self.urllib2 = urllib2

    def Log(self, txt):
        """
        Add txt to logfile history

         - *txt* a string
        """
        self.history += '%s - %s\n' % (time.strftime('%H:%M:%S'), str(txt))

    def GetLog(self):
        """
        Return the logged history
        """
        return self.history

    def Connect(self, root):
        """
        Establish HTTP connection to

         - *root* the site root "www.google.com"

        """
        try:
            self.site_root = root
            self.conn = self.urllib2.httplib.HTTPConnection(self.site_root)
            self.CONNECTED = True
            self.Log(self.site_root)
        except Exception as ex:
            print('\nConnection to {} failed!'.format(self.site_root))
            print(ex)
            self.CONNECTED = False
            self.Log('ERROR: %s' % self.site_root)
            raise RuntimeError

    def Get(self, query):
        """
        Perform an http GET using:

         - *query* e.g.
         - *reply_mode* [default=''] this is the reply mode

        For example "/semanticSBML/annotate/search.xml?q=ATP"

        """
        data1 = None
        if self.CONNECTED:
            try:
                print(query)
                HTMLhead = {'User-Agent': self.USER_AGENT}
                self.conn.request("GET", query, headers=HTMLhead)
                r1 = self.conn.getresponse()
                print(r1.status, r1.reason)
                data1 = r1.read()
                self.Log('GET %s' % (query))
            except Exception as ex:
                print('\nFailure to GET: {}{}\n'.format(self.site_root, query))
                print(ex)
                self.Log('ERROR: %s%s' % (self.site_root, query))
                raise RuntimeError
        return data1

    def Close(self):
        """
        Close the currently active connection
        """
        if self.CONNECTED:
            self.conn.close()
            self.conn = None
            self.CONNECTED = False
            self.Log('%s - connection closed' % self.site_root)
            self.site_root = None


class MIRIAMTools(object):
    """
    Tools dealing with MIRIAM annotations
    """

    def MiriamURN2IdentifiersURL(self, urn):
        urn = urn.replace('urn:miriam:', '').split(':', 1)
        urn = 'http://identifiers.org/%s/%s' % (urn[0].strip(), urn[1].strip())
        print(urn)
        return urn


class SemanticSBML(RESTClient, MIRIAMTools):
    """
    REST client for connecting to SemanticSBML services

    """

    data = None
    item_re = re.compile('<item>.+?</item>')

    def __init__(self):
        RESTClient.__init__(self)

    def quickLookup(self, txt):
        """
        Do a quick lookpup for txt using SemanticSBML (connectic if required) and return results. Returns
        a list of identifiers.org id's in descending priority (as return)

         - *txt* the string to lookup

        """

        if not self.CONNECTED:
            self.Connect("www.semanticsbml.org")

        txt = txt.strip().replace(' ', '+')
        self.data = self.Get("/semanticSBML/annotate/search.xml?q={}".format(txt))
        self.data = self.parseXMLtoText(self.data)
        return self.data

    def viewDataInWebrowser(self, maxres=10):
        """
        Attempt to view #maxres results returned by SemanticSBML in the default browser

         - *maxres* default maximum number of results to display.

        """
        cntr = 0
        for u_ in self.data:
            cntr += 1
            try:
                webbrowser.open_new_tab(u_)
            except:
                print('ERROR in url {}'.format(u_))
            if cntr >= maxres:
                print('\nMaximum results reached, set \"maxres\" to increase')
                break

    def parseXMLtoText(self, xml):
        """
        Parse the xml output by quickLookup() into a list of URL

         - *xml* XML returns from SemanticSBML

        """
        return [
            i.replace('<item>', '').replace('</item>', '').strip()
            for i in re.findall(self.item_re, xml)
        ]



"""
CBMPy: fluxmodules matroid module
=====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Arne C. Reimers
Contact email: arne.c.reimers@gmail.com
"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

from math import ceil

import time
#import sympy
import numpy
#from sympy.core.cache import clear_cache

from .sparserationals import Matrix


class Matroid:
    """A linear matroid

    It has the fields:
    elems - lists the names of the elements
    basis - lists the names of the basis elements
    nbasis - lists the names of the non-basis elements (dual basis)
    rep - reduced representation matrix,
            rows correspond to basis-elements and
            columns correspond to non-basis elements
    """
    # def __init__(self, matrix, elems):
    #    """create matroid from matrix storing elements as columns"""
    #    self.elems = __cols
    #    self.matrix = matrix
    #    self.buildRepresentation()

    def __init__(self, rep, basis, nbasis):
        """create matroid from given reduced representation matrix.

        The names of the rows are given by basis.
        The names of the columns are given by nbasis.
        """
        self.elems = basis + nbasis
        self.basis = basis
        self.nbasis = nbasis
        self.rep = rep

    def findModules(self):
        """computes the separators (modules) of this matroid

        Each module is returned as a matroid.
        """
        usedBasis = set()
        usedNBasis = set()
        modules = []
        for r in range(len(self.basis)):
            if r not in usedBasis:
                c = self.__dfs(r)
                usedBasis.update(c['basis'])
                usedNBasis.update(c['nbasis'])
#                 m = Matroid(self.rep.extract(c['basis'], c['nbasis']),
#                             [self.basis[i] for i in c['basis']],
#                             [self.nbasis[i] for i in c['nbasis']]
#                            )
                m = Matroid(self.rep[c['basis'], c['nbasis']],
                            [self.basis[i] for i in c['basis']],
                            [self.nbasis[i] for i in c['nbasis']]
                            )
                modules.append(m)
        # by construction each basis element is contained in one submatroid
        # however, this need not be the case for the non-basis elements

        # each non-basis element that we did not find is a loop
        for r in range(len(self.nbasis)):
            if r not in usedNBasis:
                #m = Matroid(sympy.zeros(0,1), [], [r])
                m = Matroid(Matrix(), [], [r])
                modules.append(m)

        # clear_cache()

        return modules

    def __dfs(self, start):
        """performs a dfs to find the connected component of basis elem start"""
        basisFound = set([start])
        nbasisFound = set()
        todo = [start]
        while len(todo) > 0:
            r = todo.pop()
            for i in range(len(self.nbasis)):
                # if r is in the cycle and the cycle has not been added,
                # then add all elements of the cycle to todo
                if (self.rep[r, i] != 0) and (i not in nbasisFound):
                    nbasisFound.add(i)
                    for s in range(len(self.basis)):
                        if (self.rep[s, i] != 0) and (s not in basisFound):
                            basisFound.add(s)
                            todo.append(s)

        return {'basis': basisFound, 'nbasis': nbasisFound}

    def dual(self):
        """ returns the dual matroid

        By this operation the original coefficient matrix is lost for the dual.
        """
        # we also care about the signs, so multiply by -1
        m = Matroid(self.rep.transpose(-1), list(self.nbasis), list(self.basis))
        return m

    def copy(self):
        """ copies this matroid """
        m = Matroid(self.rep.copy(), list(self.basis), list(self.nbasis))
        if hasattr(self, 'matrix'):
            m.matrix = self.matrix.copy()
        return m

    def contract(self, toContract):
        """ computes self / toContract

        The matroid with toContract contracted is returned.
        This instance is not modified.
        """
        # put as many elements from toDelete into basis
        self.rank(toContract)
        retain_nbasis = [i for i in range(len(self.nbasis))
                         if self.nbasis[i] not in toContract]
        retain_basis = [i for i in range(len(self.basis))
                        if self.basis[i] not in toContract]
        basis = [self.basis[i] for i in retain_basis]
        nbasis = [self.nbasis[i] for i in retain_nbasis]

        submat = self.rep[retain_basis, retain_nbasis]
        m = Matroid(submat, basis, nbasis)
        #m = self.dual().delete(toContract).dual()
        if hasattr(self, 'matrix'):
            print('TODO: do also the contraction on the matrix')

        # clear_cache()
        return m

    def exchange(self, a, b):
        """ exchanges a basis element with a non-basis element.

        requirement is that the fundamental circuit of the non-basis element
        uses the basis element
        """
        if a in self.basis:
            be = a
            assert b in self.nbasis, 'one element must be a non-basis element'
            nbe = b
        else:
            assert b in self.basis, 'one element must be a basis element'
            be = b
            assert a in self.nbasis, 'a, b must be elements of the matroid'
            nbe = a

        bidx = self.basis.index(be)
        nbidx = self.nbasis.index(nbe)
        nbidxcol = self.rep[:, nbidx]

        # We can simply switch be with nbe and only have to normalize the
        # column. This is easy, so we do it last
        div = self.rep[bidx, nbidx]
        assert div != 0, 'basis elem and non-basis elem do not share circuit'

        # There may be more columns that use the basis element
        # since the basis element changed, we have to update
        for j in range(len(self.nbasis)):
            if (j != nbidx) and (self.rep[bidx, j] != 0):
                # add self.rep[:,nbidx] to cancel
                # self.rep[bidx,j]
                coef = -self.rep[bidx, j] / div
                self.rep.coladd(j, nbidxcol, coef)
                #self.rep[:,j] += self.rep[:,nbidx]*coef
                assert self.rep[bidx, j] == 0
                # remember that we changed the meaning of the row,
                # so update the entry also
                self.rep[bidx, j] = coef

        # No do the switch with be and nbe
        self.rep[:, nbidx] /= div
        self.rep[bidx, nbidx] = 1 / div

        # and also update the column and row names
        self.basis[bidx] = nbe
        self.nbasis[nbidx] = be

    def delete(self, toDelete):
        """ computes self \ toDelete

        The matroid with toDelete deleted is returned.
        No elements of this instance are deleted. However, the current basis
        can change.
        """

        # put as many elements from toDelete into non-basis
        self.corank(toDelete)
        retain_nbasis = [i for i in range(len(self.nbasis))
                         if self.nbasis[i] not in toDelete]
        retain_basis = [i for i in range(len(self.basis))
                        if self.basis[i] not in toDelete]
        basis = [self.basis[i] for i in retain_basis]
        nbasis = [self.nbasis[i] for i in retain_nbasis]

        submat = self.rep[retain_basis, retain_nbasis]
        m = Matroid(submat, basis, nbasis)

#         #######################################################################
#         ## old, borken code
#         ######################################################################
#
#         # deletion of non-basis elements is easy
#         delete_nbasis = set([]) # columns to delete
#         for i in range(len(self.nbasis)):
#             if self.nbasis[i] in toDelete:
#                 delete_nbasis.add(i)
#
#         # for deletion of basis elements, we first have to make them non-basis
#         # elements unless they are coloops
#         delete_basis = set([]) # for coloops
#         basis_new = list(self.basis) # new basis
#         for i in range(len(self.basis)):
#             if self.basis[i] in toDelete:
#                 switch_elem = -1
#                 for j in range(len(self.nbasis)):
#                     if (self.rep[i,j] != 0) and (j not in delete_nbasis):
#                         switch_elem = j
#                 if switch_elem == -1:
#                     # i is a coloop and we can just delete it
#                     delete_basis.add(i)
#                 else:
#                     basis_new[i] = self.nbasis[switch_elem]
#                     delete_nbasis.add(switch_elem)
#                     switch_col = self.rep[:,switch_elem]
#                     # now there may be more columns that use the basis element
#                     # since the basis element changed, we have to update
#                     # We do not use self.exchange(..) because we can save some
#                     # work.
#                     for j in range(len(self.nbasis)):
#                         if (self.rep[i,j] != 0) and (j not in delete_nbasis):
#                             # add self.rep[:,switch_elem] to cancel
#                             # self.rep[i,j]
#                             coef = -self.rep[i,j] / switch_col[i,0]
#                             #coef = -self.rep[i,j] / self.rep[i,switch_elem]
#                             self.rep.coladd(j, switch_col, coef)
#                             #self.rep[:,j] += coef*self.rep[:,switch_elem]
#                             assert self.rep[i,j] == 0
#                             # remember that we changed the meaning of the row,
#                             # so update the entry also
#                             self.rep[i,j] = coef
#
#         retain_nbasis = []
#         nbasis = []
#         for i in range(len(self.nbasis)):
#             if i not in delete_nbasis:
#                 retain_nbasis.append(i)
#                 nbasis.append(self.nbasis[i])
#
#         retain_basis = []
#         basis = []
#         for i in range(len(self.basis)):
#             if i not in delete_basis:
#                 retain_basis.append(i)
#                 basis.append(basis_new[i])
#
#
#         #submat = self.rep.extract(retain_basis, retain_nbasis)
#         submat = self.rep[retain_basis, retain_nbasis]
#         m = Matroid(submat, basis, nbasis)

        # if we have the original matrix representation, also delete from that
        if hasattr(self, 'matrix'):
#             m.matrix = self.matrix.extract(
#                 range(self.matrix.rows),
#                 [i for i in range(len(self.elems))
#                     if self.elems[i] not in toDelete]
#                                           )
            m.matrix = self.matrix[
                range(self.matrix.rows()),
                [i for i in range(len(self.elems))
                    if self.elems[i] not in toDelete]
            ]
            # here the order of elements for sure gets mixed up, so set anew
            m.elems = [i for i in self.elems if i not in toDelete]

        # clear_cache()
        return m

    def parallel(self):
        """ list all groups of parallel elements

        The result is a set of frozen sets, each set listing the ids of the
        parallel elements
        """
        # basis elements are never parallel, unless there exists a
        # non-basis element with a single non-zero entry for the basis element
        # therefore, we just iterate through the non-basis elements
        par_groups = set([])

        found = set([])

        # to speed up the calculation, compute hash-values of the columns
        hashes = []
        for i in range(len(self.nbasis)):
            norm_factor = 0
            mul_factor = 73
            hashv = 1
            for j in range(len(self.basis)):
                v = self.rep[j, i]
                if v != 0:
                    if norm_factor == 0:
                        norm_factor = 1 / v
                    v = v * norm_factor
                    hashv += v.__hash__()
                hashv *= mul_factor
            hashes.append(hashv)

        for i in range(len(self.nbasis)):
            # only check if we did not already found it to be parallel to some
            # previous reaction
            if i not in found:
                par_group = [self.nbasis[i]]
                # check if parallel to some basis element
                basis_idx = -1
                basis_val = 0
                num_basis = 0
                for j in range(len(self.basis)):
                    if (num_basis < 2) and (self.rep[j, i] != 0):
                        num_basis += 1
                        if basis_idx == -1:  # first non-zero entry
                            basis_idx = j
                            basis_val = self.rep[j, i]

                if num_basis == 0:
                    # its a loop and continue
                    continue
                if num_basis == 1:
                    par_group.append(self.basis[basis_idx])

                assert basis_val != 0
                # check if parallel to other non-basis elements
                for j in range(len(self.nbasis))[i + 1:]:
                    # if two cols are the same, their hashes must be also
                    if hashes[i] == hashes[j]:
                        coef = self.rep[basis_idx, j] / basis_val
                        # check if coef != 0 not only for performance reasons
                        # this also filters out loops
                        if coef != 0:
                            diff = self.rep[:, j]
                            diff.coladd(0, self.rep[:, i], -coef)
                            if diff.isZero():
                                par_group.append(self.nbasis[j])
                                found.add(j)

                # add par_group if non-trivial
                if len(par_group) >= 2:
                    par_groups.add(frozenset(par_group))

        return par_groups

    def coparallel(self):
        """ list all groups of coparallel elements

        In metabolic network terminology the coparallel elements are the
        fully coupled reactions.

        The result is a set of frozensets, each set listing the ids of the
        coparallel elements
        """
        return self.dual().parallel()

    def isParallel(self):
        parallel = self.parallel()
        if len(parallel) != 1:
            return False

        par = parallel.pop()
        if len(par) == len(self.matroid.elems):
            return True
        else:
            return False

    def isCoparallel(self):
        coparallel = self.coparallel()
        if len(coparallel) != 1:
            return False

        par = coparallel.pop()
        if len(par) == len(self.matroid.elems):
            return True
        else:
            return False

    def fundamentalCircuit(self, e):
        """ returns the fundamental circuit of an element

        If e is a non-basis element, the fundamental circuit, i.e. the
        support of the column of e, is returned.

        If e is a basis element, the cofundamental circuit, i.e. the support of
        the row of e, is returned.

        The returned fundamental circuit does not contain e itself
        """
        fcircuit = set([])

        if e in self.basis:
            # return cofundamental circuit (row)
            idx = self.basis.index(e)
            for j in range(len(self.nbasis)):
                if self.rep[idx, j] != 0:
                    fcircuit.add(self.nbasis[j])
        else:
            assert e in self.nbasis, 'this works only for elements of the matroid'
            # return fundamental circuit (column)
            idx = self.nbasis.index(e)
            for j in range(len(self.basis)):
                if self.rep[j, idx] != 0:
                    fcircuit.add(self.basis[j])
        return fcircuit

    def fundamentalCircuitValues(self, e):
        """ returns the fundamental circuit of an element with the coefficients

        If e is a non-basis element, the fundamental circuit, i.e. the
        support of the column of e, is returned.

        If e is a basis element, the cofundamental circuit, i.e. the support of
        the row of e, is returned.

        The returned fundamental circuit does not contain e itself. If needed,
        it should be added manually afterwards with a coefficient of 1.
        """
        fcircuit = {}

        if e in self.basis:
            # return cofundamental circuit (row)
            idx = self.basis.index(e)
            for j in range(len(self.nbasis)):
                if self.rep[idx, j] != 0:
                    fcircuit[self.nbasis[j]] = self.rep[idx, j]
        else:
            assert e in self.nbasis, 'this works only for elements of the matroid'
            # return fundamental circuit (column)
            idx = self.nbasis.index(e)
            for j in range(len(self.basis)):
                if self.rep[j, idx] != 0:
                    fcircuit[self.basis[j]] = self.rep[j, idx]
        return fcircuit

    def rank(self, elems):
        """ computes the rank of elems

        This method puts as many elements of elems into the basis. The number of
        elements from elems in the final basis is returned.
        """
        # try to make as many nbasis elements of elems basis-elements
        for i in range(len(self.nbasis)):
            if self.nbasis[i] in elems:
                # make it a basis element
                for j in range(len(self.basis)):
                    if (self.rep[j, i] != 0) and (self.basis[j] not in elems):
                        self.exchange(self.basis[j], self.nbasis[i])
                        break

        # count basis elements in elems
        numBasis = 0
        for i in range(len(self.basis)):
            if self.basis[i] in elems:
                numBasis += 1
        return numBasis

    def corank(self, elems):
        """ computes the corank of elems

        This method puts as many elements of elems into the non-basis. The
        number of elements from elems in the final non-basis is returned.
        """
        # try to make as many nbasis elements of elems basis-elements
        for i in range(len(self.basis)):
            if self.basis[i] in elems:
                # make it a nbasis element
                for j in range(len(self.nbasis)):
                    if (self.rep[i, j] != 0) and (self.nbasis[j] not in elems):
                        self.exchange(self.nbasis[j], self.basis[i])
                        break

        # count non-basis elements in elems
        numNBasis = 0
        for i in range(len(self.nbasis)):
            if self.nbasis[i] in elems:
                numNBasis += 1
        return numNBasis

    def isIndependent(self, elems):
        """ checks if the given set of elements is independent.

        This does not modify this matroid, but may change its internal basis.
        If this method returns true, the current basis contains all elements of
        elems
        """
        # no independent set can have more elements than any basis
        if len(elems) > len(self.basis):
            return False

        # we check by making all elements of elems basis elements
        for i in range(len(self.nbasis)):
            if self.nbasis[i] in elems:
                # make it a basis element
                found = False
                for j in range(len(self.basis)):
                    if (self.rep[j, i] != 0) and (self.basis[j] not in elems):
                        self.exchange(self.basis[j], self.nbasis[i])
                        found = True
                        break
                if not found:
                    return False
        # all elements of elems are now basis elements
        return True

    def isCoIndependent(self, elems):
        """ checks if the given set of elements is coindependent.

        This does not modify this matroid, but may change its internal basis.
        If this method returns true, the current basis contains all no elements
        of elems

        This is the same as isIndependent, only on the dual
        """
        # no independent set can have more elements than any basis
        if len(elems) > len(self.nbasis):
            return False

        # we check by making all elements of elems non-basis elements
        for j in range(len(self.basis)):
            if self.basis[j] in elems:
                # make it a basis element
                found = False
                for i in range(len(self.nbasis)):
                    if (self.rep[j, i] != 0) and (self.nbasis[i] not in elems):
                        self.exchange(self.basis[j], self.nbasis[i])
                        found = True
                        break
                if not found:
                    return False
        # all elements of elems are now non-basis elements
        return True

    def connectivity(self, separation):
        """ evaluate the connectivity function for the given separation.

            separation is a set of elements encoding one side of the separation.
            The other side is implied.
        """
        # for the computation we do the following:
        # put as many elements as possible from separation into the basis
        # this gives us the rank of separation.
        # We do the same for the complement.

        # make a copy, to reduce the number of pivots
        copy = self.copy()
        r1 = self.rank(separation)
        r2 = copy.rank(set(self.elems).difference(separation))
        r = len(self.basis)
        # clear_cache()
        return r1 + r2 - r

    def getInterface(self, separation):
        """ compute a list of circuits that define the interface of the
        given separation """

        # theoretically the following would be sufficient:
        # self.rank(separation)
        # but we do something more complicated...

        # Before we start the computation, we observe that eventually it is good
        # to select circuits that are as simple as possible, so that the
        # selected interface vectors become as simple as possible.
        # typically the biomass reactions induce very_ugly circuits, hence we
        # do not want them to be in the basis, because then the fundamental
        # circuits get ugly.
        # HACK: We identify biomass reactions by their reaction NAME!!!
        # TODO: get some cleaner solution for the future!

        # find non-biomass reactions in the separation
        separation_non_bio = [r for r in separation if r.lower().find('bio') == -1]
        # find biomass reactions in the separation
        separation_bio = [r for r in separation if r.lower().find('bio') != -1]

        assert len(separation_non_bio) + len(separation_bio) == len(separation)

        self.rank(separation)

        # try to turn as many non-bio nbasis elements into basis elements by
        # exchanging with bio elements
        # use the property that every fundamental circuit of nbasis separation
        # elements only contains separation elements
        for i in range(len(self.nbasis)):
            if self.nbasis[i] in separation and self.nbasis[i].find('bio') == -1:
                # make it a basis element
                for j in range(len(self.basis)):
                    if (self.rep[j, i] != 0) and (self.basis[j].lower().find('bio') != -1):
                        assert self.basis[j] in separation
                        self.exchange(self.basis[j], self.nbasis[i])
                        break

#        print(separation_bio)
#        self.rank(separation_bio)
#        print(self.basis)
#        print(separation_non_bio)
#        self.rank(separation_non_bio)
#        print(self.basis)
        # a maximal amount of separation elements are in the basis and of that
        # as few as possible are biomass reactions

        # Verify this!
        for r in self.nbasis:
            if r in separation:
                c = self.fundamentalCircuit(r)
                for s in c:
                    assert s in separation, (r, c, separation)

        # find biomass reactions not in the separation
        coseparation_bio = [r for r in self.basis if r not in separation and r.lower().find('bio') != -1]
        # ideally these reactions should not be in the basis, so try to
        # put a maximal amout of them into the nbasis
#        print(coseparation_bio)
        self.corank(coseparation_bio)
#        print(self.basis)
        # still the property is satisfied that a maximal amount of separation
        # elments are in the basis

        # we now have the following properties:
        # every fundamental circuit of an nbasis element in separation uses
        # only basis elements in separation
        # every cofundamental cocircuit of a basis element outside separation
        # uses only non-basis elements outside separation

        # Verify this!
        for r in self.nbasis:
            if r in separation:
                c = self.fundamentalCircuit(r)
                for s in c:
                    assert s in separation

        # we only have to deal with the fundamental circuits of nbasis elements
        # outside of separation.
        # They already define the interface, but the representation might be
        # redundant
        # Hence, we want to minimize it

        # Since the basis elements outside of the separation are of no interest,
        # we would simply contract them. However, this would make it more
        # difficult to reconstruct the original circuits.
        # There, we manually extract the reduced representation and take the
        # submatrix with rows of elements in separation and columns not in
        # separation
        rows = [i for i in range(len(self.basis)) if self.basis[i] in separation]
        cols = [i for i in range(len(self.nbasis)) if self.nbasis[i] not in separation]
        colNames = [self.nbasis[i] for i in cols]

        submat = self.rep[rows, cols]
        # redundant circuits are linearly dependent on the existing circuits
        # hence, we just need to find a basis of the reduced matrix
        # this we do by simply building the matroid
        mred = fromMatrix(submat, colNames)

        # here we continue our hack about the biomass reactions.
        # Theoretically, all bases of mred are equally good, but since we want
        # to have simple circuits, we prefer to have no biomass reactions in
        # the selected circuits and thus not in the basis of mred
#        print(mred.basis)
        col_bio = [r for r in colNames if r.lower().find('bio') != -1]
        mred.corank(col_bio)
#        print(mred.basis)

        interface = []
        for c in mred.basis:
            circuit = self.fundamentalCircuitValues(c)
            circuit[c] = 1  # the circuit originally does not contain
            interface.append(circuit)

        assert self.connectivity(separation) == len(interface)
        return interface

    def similarityMatrix(self, stype):
        """ computes a similarity matrix on the elements

        For each pair of elements the similarity matrix has an entry that
        gives the similarity.

        stype gives what kind of similarity matrix should be computed.
        Possibilities are:
        direct - direct from the columns of the original coefficient matrix
            (needs the original coefficient matrix and in the case of
            contractions the result is not well defined)
        __cols - normalized from the columns (rows are orthogonal)
        rows - normalized from the rows (columns are orthogonal)
        combined - takes max(__cols, rows)
        """
        if stype == "direct":
            return self.__directSimilarityMatrix()
        elif stype == "cols":
            return self.__colSimilarityMatrix()
        elif stype == "rows":
            return self.__rowSimilarityMatrix()
        elif stype == "combined":
            rowSim = self.__rowSimilarityMatrix()
            colSim = self.__colSimilarityMatrix()
            return rowSim.max(colSim)
        else:
            print('unknown option')
            assert False

    def __directSimilarityMatrix(self):
        mat = self.matrix.toNumpy()

        elems = self.elems

        # compute similarity matrix
        similarity = numpy.zeros((len(elems), len(elems)))
        # precompute norms
        norms = numpy.zeros(len(elems))
        for i in range(len(elems)):
            norms[i] = numpy.sqrt(numpy.dot(mat[:, i], mat[:, i]))
        printMod = ceil(1.0e9 / (len(elems)**2))
        for i in range(len(elems)):
            if i % printMod == 0:
                print('building similarity matrix: %d of %d' % (i, len(elems)))
            for j in range(len(elems)):
                if i != j:
                    similarity[i, j] = numpy.abs(numpy.dot(mat[:, j], mat[:, i]))
                    similarity[i, j] /= norms[i] * norms[j]
                else:
                    similarity[i, j] = -1

        return SimilarityMatrix(similarity, elems)

    def __rowSimilarityMatrix(self):
        mat = self.rep.toNumpy()
        #mat = sympy.matrix2numpy(self.rep)
        mat = numpy.concatenate((numpy.identity(len(self.nbasis)), mat))

        elems = self.nbasis + self.basis

        printMod = ceil(1.0e9 / (len(elems)**2))
        # GramSchmidt orthonornalization
        for i in range(len(self.nbasis)):
            if i % printMod == 0 and i > 0:
                print('normalizing matrix: %d of %d' % (i, len(elems)))
            for j in range(i - 1):
                mat[:, i] -= numpy.dot(mat[:, j], mat[:, i]) * mat[:, j]
                assert numpy.abs(numpy.dot(mat[:, i], mat[:, j])) < 1e-7
            mat[:, i] /= numpy.sqrt(numpy.dot(mat[:, i], mat[:, i]))
            assert abs(numpy.dot(mat[:, i], mat[:, i]) - 1) < 1e-7

        # compute similarity matrix
        similarity = numpy.zeros((len(elems), len(elems)))
        # precompute norms
        norms = numpy.zeros(len(elems))
        for i in range(len(elems)):
            norms[i] = numpy.sqrt(numpy.dot(mat[i, :], mat[i, :]))
        for i in range(len(elems)):
            if i % printMod == 0 and i > 0:
                print('building similarity matrix: %d of %d' % (i, len(elems)))
            for j in range(len(elems)):
                if i != j:
                    similarity[i, j] = numpy.abs(numpy.dot(mat[j, :], mat[i, :]))
                    similarity[i, j] /= norms[i] * norms[j]
                else:
                    similarity[i, j] = -1

        return SimilarityMatrix(similarity, elems)

    def __colSimilarityMatrix(self):
        # we cannot just reuse __rowSimilarityMatrix, because that would
        # give a different order of the elements
        mat = self.rep.toNumpy()
        #mat = sympy.matrix2numpy(self.rep)
        mat = numpy.concatenate((mat, numpy.identity(len(self.basis))), 1)

        elems = self.nbasis + self.basis

        printMod = ceil(1.0e9 / (len(elems)**2))
        # GramSchmidt orthonornalization
        for i in range(len(self.basis)):
            if i % printMod == 0 and i > 0:
                print('normalizing matrix: %d of %d' % (i, len(elems)))
            for j in range(i - 1):
                mat[i, :] -= numpy.dot(mat[j, :], mat[i, :]) * mat[j, :]
            mat[i, :] /= numpy.sqrt(numpy.dot(mat[i, :], mat[i, :]))

        # compute similarity matrix
        similarity = numpy.zeros((len(elems), len(elems)))
        # precompute norms
        norms = numpy.zeros(len(elems))
        for i in range(len(elems)):
            norms[i] = numpy.sqrt(numpy.dot(mat[:, i], mat[:, i]))
        for i in range(len(elems)):
            if i % printMod == 0 and i > 0:
                print('building similarity matrix: %d of %d' % (i, len(elems)))
            for j in range(len(elems)):
                if i != j:
                    similarity[i, j] = numpy.abs(numpy.dot(mat[:, j], mat[:, i]))
                    similarity[i, j] /= norms[i] * norms[j]
                else:
                    similarity[i, j] = -1

        return SimilarityMatrix(similarity, elems)


def fromMatrix(matrix, cols):
    """ builds the reduced representation matrix

        Given a matrix of columns (self.matrix) it builds the reduced
        representation matrix (self.rep) that has rows for each basis
        element and columns for each non-basis element.

        From this form the fundamental cycles can easily be read of and
        other operations like deletion and contraction are easy.
    """
    assert matrix.cols() <= len(cols), 'for each column a name is needed'
    inv_list = []  # basis vectors in (transformed) triangular form
    row_perm = []  # stores which index we use for each basis element
    # inv_cycle is matrix to reconstruct basis vectors,
    # used to compute cycles for non-basis elements
    #inv_cycle = sympy.zeros(len(cols))
    inv_cycle = Matrix()
    basis = []  # names of basis elements
    nbasis = []  # names of non-basis elements
    repRawData = []  # raw form of representation matrix

    time_build_inv_cycle_col = 0
    time_find_row = 0
    time_add_dependent = 0
    time_add_basis = 0

    if len(cols) >= 100:
        print("creating matroid representation: ", end="")
    for i in range(len(cols)):
        if i > 0 and i % 100 == 0:
            print(i, end=" ")
        c = matrix[:, i]
        # eliminate existing coordinates (in row_perm)
        stime = time.time()
        res = c.copy()  # residual
        # in case its linearly independent, store the linear transformation
        #inv_cycle_col = sympy.zeros(len(cols),c=1)
        inv_cycle_col = Matrix()
        inv_cycle_col[len(row_perm), 0] = 1
        for j in range(len(row_perm)):
            coef = c[row_perm[j], 0]
            if coef != 0:
                # update residual
                res.coladd(0, inv_list[j], -coef)
                #res -= inv_list[j]*coef
                inv_cycle_col.coladd(0, inv_cycle[:, j], -coef)
                #inv_cycle_col -= inv_cycle[:,j]*coef

        time_build_inv_cycle_col += time.time() - stime
        stime = time.time()
        # check if res contains non-zero entries and
        # if yes, find entry with minimum no. of non-zeros in other __cols

        min_index = -1
        min_count = -1
        for j in range(matrix.rows()):
            if res[j, 0] != 0:
                # print(row_perm)
                # print(res)
                assert j not in row_perm
                # count no. non-zeros in other __cols
                count = 0
                for x in inv_list:
                    if x[j, 0] != 0:
                        count += 1
                # update min_count and min_index
                if min_index < 0 or count < min_count:
                    min_index = j
                    min_count = count

        time_find_row += time.time() - stime
        stime = time.time()
        if min_index == -1:
            # the column i is linearly dependent
            # therefore, add fundamental circuit and add non-basis element

            # reconstruct linear combination of original vectors (cycle)
            sub_inv = inv_cycle[0:len(basis), 0:len(basis)]
#            v = sub_inv*c.extract(row_perm,[0])
            v = sub_inv * c[row_perm, [0]]

            # new non-basis element
            nbasis.append(cols[i])
            # corresponding cycle
            repRawData.append(-v)
            time_add_dependent += time.time() - stime
        else:
            # the column i is linearly independent of other columns
            # therefore, add to basis
            basis.append(cols[i])

            # normalize such that the new basis-vector has entry 1 in row j
            # we choose min_index as the 1-entry such that we minimize the
            # number of transformations
            div = res[min_index, 0]
            inv_cycle_col /= div
            # store transformation
            inv_cycle.coladd(len(basis) - 1, inv_cycle_col)
            #inv_cycle[:,len(basis)-1] = inv_cycle_col
            res /= div

            # eliminate non-zero coefficients of existing basis-vectors
            # in row min_index
            for k in range(len(inv_list)):
                x = inv_list[k]
                if x[min_index, 0] != 0:
                    # also update the transformation matrix
                    inv_cycle.coladd(k, inv_cycle_col, -x[min_index, 0])
                    #inv_cycle[:,k] -= inv_cycle_col*x[min_index,0]
                    x.coladd(0, res, -x[min_index, 0])
                    #inv_list[k] -= res*x[min_index,0]

            # add basis-vector
            inv_list.append(res)
            row_perm.append(min_index)
            time_add_basis += time.time() - stime

    if len(cols) >= 100:
        print("")  # to finish the line

#    print("time build inv cycle col: %s" % time_build_inv_cycle_col)
#    print("time find row %s" % time_find_row)
#    print("time add dependent %s" % time_add_dependent)
#    print("time add basis %s" % time_add_basis)
    # finally, build representation matrix
    #rep = sympy.zeros(len(basis), len(nbasis))
    rep = Matrix()
    for i in range(len(repRawData)):
        rep[:, i] = repRawData[i]  # TODO: I think this can be greatly improved!

    m = Matroid(rep, basis, nbasis)
    m.matrix = matrix
    m.elems = cols  # just make sure that the order is correct
    # clear_cache()
    return m


class SimilarityMatrix:

    def __init__(self, matrix, elems):
        """ builds a similarity matrix out of a matrix and element names.

        the matrix must be square and have as many rows/columns as elems has
        entries
        """
        self.matrix = matrix
        self.elems = elems

    def max(self, simmatrix):
        assert simmatrix.elems == self.elems
        mat = numpy.maximum(self.matrix, simmatrix.matrix)

        return SimilarityMatrix(mat, self.elems)

    def __getitem__(self, i):
        return self.matrix[i]

    def __setitem__(self, i, v):
        self.matrix[i] = v



"""
CBMPy: fluxmodules sparse rationals module
=====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Arne C. Reimers
Contact email: arne.c.reimers@gmail.com
"""


from fractions import Fraction
#from .myfraction import Fraction
import numpy


class Matrix:
    """ sparse matrix of rational numbers (fractions)

    The matrix grows by demand.
    """

    def __init__(self, listRep=None):
        """" Creates empty matrix.
        """
        self.__cols = []
        self.__nrows = 0

        if listRep != None:
            self.addListRep(listRep)

    def rows(self):
        return self.__nrows

    def cols(self):
        return len(self.__cols)

    def getCol(self, idx):
        """ returns the column at index idx as a frozen set of key-value pairs

        The returned dictionary does not contain any superfluous zeros.
        """
        out = set()
        # sanitize column idx
        if len(self.__cols) <= idx:
            return frozenset()
        else:
            c = self.__cols[idx]
            for k, v in c.items():
                if v == 0:
                    del c[k]
                else:
                    out.add((k, v))
            return frozenset(out)

    def __getitemIdx(self, source, rowidx, out):
        for c in source:
            col = {}
            for r, v in c.items():
                i = rowidx(r)
                if i != None:
                    col[i] = v
            out.__cols.append(col)

    def transpose(self, scalar=1):
        """ creates a transpose of the matrix

        Optionally, the values can be multiplied by a scalar.
        This is for example used to compute the dual matroid.
        """
        assert self.checkRows()
        out = Matrix()
        for i in range(self.rows()):
            out.__cols.append({})
        for ci, c in enumerate(self.__cols):
            for r, v in c.items():
                out.__cols[r][ci] = v * scalar
#        out.__cols[r] = [self.__cols[c][r] for c in range(self.cols()) if self.__cols[c].has_key(r)]
#        [out.__cols[r][ci]=v for  ci, c in enumerate(self.__cols) for r,v in c.items()]
        out.__nrows = len(self.__cols)
        assert out.checkRows()
        return out

    def copy(self):
        """ copies this matrix """
        assert self.checkRows()
        out = Matrix()
        for c in self.__cols:
            c2 = c.copy()
            out.__cols.append(c2)
        out.__nrows = self.__nrows
        assert out.checkRows()
        return out

    def __getitem__(self, indices):
        """ fetch entries of this matrix

        indices must be a 2-tuple indexing rows and columns.
        This returns a matrix.

        If the index is a 2-tuple of ints, a Fraction is returned.
        Otherwise, a Matrix is returned.
        """
        assert len(indices) == 2
        assert self.checkRows()
        rows = indices[0]
        cols = indices[1]

        if isinstance(cols, int):
            if isinstance(rows, int):
                if cols < self.cols():
                    return self.__cols[cols].get(rows, 0)
                else:
                    return 0
            else:
                cols = [cols]
        elif isinstance(rows, int):
            rows = [rows]

        if isinstance(rows, slice):
            rlen = len(xrange(*rows.indices(self.rows())))
        else:
            rlen = len(rows)

        out = Matrix()
        out.__nrows = rlen

        if isinstance(rows, slice):
            ind = rows.indices(self.rows())
            def idx(x): return revslice(x, ind)
        else:
            ind = {}
            for i, v in enumerate(rows):
                ind[v] = i

            def idx(x): return ind.get(x, None)

        if isinstance(cols, slice):
            subcols = self.__cols[cols]
        else:
            subcols = [self.__cols[i] for i in cols if i < self.cols()]

        self.__getitemIdx(subcols, idx, out)
        assert out.checkRows()
        return out

    def __resetSlice(self, source, rslice):
        assert self.checkRows()
        for c in source:
            for r in c.keys():
                if (r >= rslice.start and
                        r < rslice.stop and
                        r - rslice.start % rslice.step == 0):
                    del c[r]
        assert self.checkRows()

    def __resetTest(self, source, rTest):
        assert self.checkRows()
        for c in source:
            for r in c.keys():
                if r in rTest:
                    del c[r]
        assert self.checkRows()

    def reset(self, rows, cols):
        """ reset all elements in rows x __cols to zero

        This is equivalent (but a bit faster) to self[rows, __cols] = 0
        """
        assert self.checkRows()
        if isinstance(cols, slice):
            subcols = self.__cols[cols]
        else:
            subcols = [self.__cols[i] for i in cols if i < self.cols()]

        if isinstance(rows, slice):
            self.__resetSlice(subcols, rows)
        else:
            self.__resetTest(subcols, rows)
        assert self.checkRows()

    def __iadd__(self, other):
        """ += operation """
        assert self.checkRows()
        assert other.checkRows()
        mcol = other.cols()
        while self.cols() < mcol:
            self.__cols.append({})
        # copy new values
        for i, c in enumerate(other.__cols):
            mycol = self.__cols[i]
            for r, v in c.items():
                nv = mycol.get(r, 0) + v
                if nv != 0:
                    mycol[r] = nv
                elif mycol.has_key(r):
                    del mycol[r]

        self.__nrows = max(self.__nrows, other.__nrows)
        assert self.checkRows()
        return self

    def __isub__(self, other):
        """ -= operation """
        assert self.checkRows()
        assert other.checkRows()
        mcol = other.cols()
        while self.cols() < mcol:
            self.__cols.append({})
        # copy new values
        for i, c in enumerate(other.__cols):
            mycol = self.__cols[i]
            for r, v in c.items():
                nv = mycol.get(r, 0) - v
                if nv != 0:
                    mycol[r] = nv
                elif mycol.has_key(r):
                    del mycol[r]

        self.__nrows = max(self.__nrows, other.__nrows)
        assert self.checkRows()
        return self

    def __idiv__(self, scalar):
        """ /= operation """
        assert self.checkRows()
        for c in self.__cols:
            for r, v in c.items():
                c[r] = v / scalar
        return self

    def __itruediv__(self, scalar):
        return self.__idiv__(scalar)

    def __imul__(self, v):
        """ *= operation

        works for scalars and matrices.
        """
        assert self.checkRows()
        if isinstance(v, Matrix) and (v.rows() > 1 or v.cols() > 1):
            # proper matrix multiplication
            # we do not care about matching matrix dimensions, because
            # we can always fill up with zeros
            out = Matrix()
            for tr in range(self.rows()):
                for ci, c in enumerate(v.__cols):
                    x = 0
                    for ri, rv in c.items():
                        x += self[tr, ri] * rv
                    out[tr, ci] = x
            # copy result to self
            self.__cols = out.__cols
            # the number of rows stays the same so no update needed
        else:
            # just mutliplication with a scalar
            if isinstance(v, Matrix):
                scalar = v[0, 0]
            else:
                scalar = v
            for c in self.__cols:
                for r, val in c.items():
                    c[r] = val * scalar
        return self

    def __add__(self, other):
        """ + operation """
        assert self.checkRows()
        out = Matrix()
        out.__iadd__(self)
        out.__iadd__(other)
        return out

    def __sub__(self, other):
        """ - operation """
        assert self.checkRows()
        out = Matrix()
        out.__iadd__(self)
        out.__isub__(other)
        return out

    def __mul__(self, v):
        """ * operation (for multiplication with scalars)"""
        assert self.checkRows()
        if isinstance(v, Matrix) and (v.rows() > 1 or v.cols() > 1):
            # proper matrix multiplication
            # we do not care about matching matrix dimensions, because
            # we can always fill up with zeros
            out = Matrix()
            for tr in range(self.rows()):
                for ci, c in enumerate(v.__cols):
                    x = 0
                    for ri, rv in c.items():
                        x += self[tr, ri] * rv
                    out[tr, ci] = x
        else:
            # just mutliplication with a scalar
            if isinstance(v, Matrix):
                scalar = v[0, 0]
            else:
                scalar = v
            out = Matrix()
            for c in self.__cols:
                oc = {}
                for r, val in c.items():
                    oc[r] = val * scalar
                out.__cols.append(oc)
            out.__nrows = self.__nrows
#        assert self.checkRows()
#        out = Matrix()
#        out.__iadd__(self)
#        out.__imul__(v)
        return out

    def __div__(self, scalar):
        """ / operation (for division by scalars)"""
        assert self.checkRows()
        out = Matrix()
        out.__iadd__(self)
        out.__idiv__(scalar)
        return out

    def coladd(self, col, v, scalar=1):
        """ computes self[:,col] += scalar * v

        col can only be a single index
        """
        assert self.checkRows()
        assert v.checkRows()
        assert v.cols() >= 1

        while self.cols() <= col:
            self.__cols.append({})

        c = self.__cols[col]
        oc = v.__cols[0]
        for r, val in oc.items():
            nv = c.get(r, 0) + scalar * val
            if nv != 0:
                c[r] = nv
                if r >= self.__nrows:
                    self.__nrows = r + 1
            else:
                del c[r]

        assert self.checkRows()

    def __setitem__(self, indices, v):
        assert len(indices) == 2
        assert self.checkRows()
        rows = indices[0]
        cols = indices[1]

        if isinstance(cols, int):
            cols = [cols]
        if isinstance(rows, int):
            rows = [rows]

        if isinstance(rows, slice):
            if rows.start == None:
                rstart = 0
            else:
                rstart = rows.start
            if rows.step == None:
                maxrow = max(self.rows(), rstart + v.rows())
            elif rows.step > 0:
                maxrow = max(self.rows(), rstart + v.rows() * rows.step)
            else:
                maxrow = self.rows()
            rows = xrange(*rows.indices(maxrow))
        if isinstance(cols, slice):
            if cols.start == None:
                cstart = 0
            else:
                cstart = cols.start
            if cols.step == None:
                maxcol = max(self.cols(), cstart + v.cols())
            elif cols.step > 0:
                maxcol = max(self.cols(), cstart + v.cols() * cols.step)
            else:
                maxcol = self.cols()
            cols = xrange(*cols.indices(maxcol))

        if len(rows) == 1 and len(cols) == 1:
            row = rows[0]
            col = cols[0]
            # just set single value
            if v == 0:
                if self.cols() > col:
                    if self.__cols[col].has_key(row):
                        del self.__cols[col][row]
            else:
                while self.cols() <= col:
                    self.__cols.append({})
                self.__nrows = max(self.__nrows, row + 1)
                if isinstance(v, Matrix):
                    assert v.cols() <= 1
                    assert v.rows() <= 1
                    if v.cols() == 0 or v.rows() == 0:
                        if self.__cols[col].has_key(row):
                            del self.__cols[col][row]
                    else:
                        self.__cols[col][row] = v[0, 0]
                elif isinstance(v, Fraction):
                    self.__cols[col][row] = v
                else:
                    self.__cols[col][row] = Fraction(v)  # make sure its a fraction

        else:
            assert isinstance(v, Matrix)
            assert self.checkRows()
            # empty matrix
            self.reset(rows, cols)
            # add empty columns if necessary
            mcol = max(cols)
            while self.cols() <= mcol:
                self.__cols.append({})
            # copy new values
            for i, c in enumerate(v.__cols):
                for r, val in c.items():
                    assert i < len(cols)
                    assert r < len(rows), "%d < %d %d" % (r, len(rows), v.rows())
                    self.__cols[cols[i]][rows[r]] = val
                    self.__nrows = max(self.__nrows, rows[r] + 1)
        assert self.checkRows()

    def __neg__(self):
        """ return -self """
        out = Matrix()
        out.__isub__(self)
        return out

    def __eq__(self, other):
        if isinstance(other, Matrix):
            for i in range(max(len(self.__cols), len(other.__cols))):
                if i >= len(self.__cols):
                    # check if the other column only contains zeros
                    for v in other.__cols[i].values():
                        if v != 0:
                            return False
                elif i >= len(other.__cols):
                    # check if the own column only contains zeros
                    for v in self.__cols[i].values():
                        if v != 0:
                            return False
                else:
                    # is a valid index for both matrices
                    sc = self.__cols[i]
                    oc = other.__cols[i]
                    for k, v in sc.items():
                        if oc[k] != v:
                            return False
                    for k, v in oc.items():
                        if sc[k] != v:
                            return False
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def addMetabolicNetwork(self, cmod):
        """ adds the stoichiometric matrix of a metabolic network

        This is the default method to load the stoichiometric matrix into
        ths matrix. Just create an empty matrix and call this function.

        A list of column names is returned.
        """
        var_spec_id = [s.getId() for s in cmod.species if not s.is_boundary]
        num_col = len(cmod.reactions)
        while self.cols() < num_col:
            self.__cols.append({})
        labels = []
        for c in range(num_col):
            labels.append(cmod.reactions[c].getId())
            for reag in cmod.reactions[c].getStoichiometry():
                if reag[1] in var_spec_id:
                    r = var_spec_id.index(reag[1])
                    self.__cols[c][r] = self.__cols[c].get(r, 0) + Fraction(reag[0])
        self.__nrows = len(var_spec_id)
        assert self.checkRows()
        return labels, var_spec_id

    def addListRep(self, matrix):
        """ adds matrix in list representation (ordered by rows) """
        self.__nrows = max(self.__nrows, len(matrix))
        for ri, r in enumerate(matrix):
            while len(r) > self.cols():
                self.__cols.append({})
            for ci, v in enumerate(r):
                if v != 0:
                    col = self.__cols[ci]
                    col[ri] = col.get(ri, 0) + Fraction(v)

        assert self.checkRows()

    def toNumpy(self):
        """ returns floating point numpy representation.

        The numpy matrix is a full matrix!
        """
        assert self.checkRows()
        # find max row
        out = numpy.zeros((self.rows(), self.cols()))
        for i, c in enumerate(self.__cols):
            for r, v in c.items():
                out[r, i] = v
        return out

    def checkRows(self):
#        for c in self.__cols:
#            for r in c.keys():
#                assert r < self.rows()
#                if r >= self.rows():
#                    return False
        return True

    def isZero(self):
        for c in self.__cols:
            for r in c.values():
                if r != 0:
                    return False
        return True


def revslice(x, sl):
    if sl[2] == 0:
        return None
    elif sl[2] > 0:
        if x >= sl[1]:
            return None
        if x < sl[0]:
            return None
    else:
        # sl[2] < 0
        if x >= sl[0]:
            return None
        if x < sl[1]:
            return None
    y = x - sl[0]
    if y % sl[2] != 0:
        return None
    return y / sl[2]



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import

# from __future__ import unicode_literals


from cbmpy.CBConfig import __CBCONFIG__ as __CBCONFIG__

__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']

import time
import os
import scipy
import scipy.linalg

if int(scipy.__version__.split('.')[1]) < 12:
    myfblas = scipy.linalg.fblas
    myflapack = scipy.linalg.flapack
else:
    myfblas = scipy.linalg.blas
    myflapack = scipy.linalg.lapack

import copy

__doc__ = """
          PyscesStoich
          ------------
          PySCeS stoichiometric analysis classes.

          """

##  print('Stoichiometry ver ' + __version__ + ' runtime: '+ time.strftime("%H:%M:%S"))

__psyco_active__ = 0
# try:
##  import psyco
# psyco.profile()
##  __psyco_active__ = 1
# print 'PySCeS Stoichiometry Module is now PsycoActive!'
# except:
##  __psyco_active__ = 0

# stoich_zero_valM = scipy.machar.MachAr().eps*2.0e4 # safer --> about 4e-11 (unofficially - a minpivot size)
# stoich_zero_valM = scipy.machar.MachAr().eps*2.0e4 # safe --> about 4e-12 (unofficially - a minpivot size)
# print '\tStoichiometric precision = ', stoich_zero_valM


class StructMatrix:
    """
    This class is specifically designed to store structural matrix information
    give it an array and row/col index permutations it can generate its own
    row/col labels given the label src.
    """

    array = None
    ridx = None
    cidx = None
    row = None
    col = None
    shape = None

    def __init__(self, array, ridx, cidx, row=None, col=None):
        """
        Instantiate with array and matching row/col index arrays, optional label arrays
        """
        self.array = array
        self.ridx = ridx
        self.cidx = cidx
        self.row = row
        self.col = col
        self.shape = array.shape

    def __call__(self):
        return self.array

    def getRowsByIdx(self, *args):
        """Return the rows referenced by index (1,3,5)"""
        return self.array.take(args, axis=0)

    def getColsByIdx(self, *args):
        """Return the columns referenced by index (1,3,5)"""
        return self.array.take(args, axis=1)

    def setRow(self, src):
        """
        Assuming that the row index array is a permutation (full/subset)
        of a source label array by supplying that source to setRow it
        maps the row labels to ridx and creates self.row (row label list)
        """
        self.row = [src[r] for r in self.ridx]

    def setCol(self, src):
        """
        Assuming that the col index array is a permutation (full/subset)
        of a source label array by supplying that src to setCol
        maps the row labels to cidx and creates self.col (col label list)
        """

        self.col = [src[c] for c in self.cidx]

    def getRowsByName(self, *args):
        """Return the rows referenced by label ('s','x','d')"""
        assert self.row != None, "\nI need row labels"
        try:
            return self.array.take([self.row.index(l) for l in args], axis=0)
        except Exception as ex:
            print(ex)
            print("\nValid row labels are: %s" % self.row)
            return None

    def getColsByName(self, *args):
        """Return the columns referenced by label ('s','x','d')"""
        assert self.col != None, "\nI need column labels"
        try:
            return self.array.take([self.col.index(l) for l in args], axis=1)
        except Exception as ex:
            print(ex)
            print("Valid column labels are: %s" % self.col)
            return None

    def getLabels(self, axis='all'):
        """Return the matrix labels ([rows],[cols]) where axis='row'/'col'/'all'"""
        if axis == 'row':
            return self.row
        elif axis == 'col':
            return self.col
        else:
            return self.row, self.col

    def getIndexes(self, axis='all'):
        """Return the matrix indexes ([rows],[cols]) where axis='row'/'col'/'all'"""
        if axis == 'row':
            return self.ridx
        elif axis == 'col':
            return self.cidx
        else:
            return self.ridx, self.cidx

    def getByIdx(self, row, col):
        assert row in self.ridx, '\n%s is an invalid index' % row
        assert col in self.cidx, '\n%s is an invalid index' % col
        return self.array[row, col]

    def getByName(self, row, col):
        assert row in self.row, '\n%s is an invalid name' % row
        assert col in self.col, '\n%s is an invalid name' % col
        return self.array[self.row.index(row), self.col.index(col)]

    def setByIdx(self, row, col, val):
        assert row in self.ridx, '\n%s is an invalid index' % row
        assert col in self.cidx, '\n%s is an invalid index' % col
        self.array[row, col] = val

    def setByName(self, row, col, val):
        assert row in self.row, '\n%s is an invalid name' % row
        assert col in self.col, '\n%s is an invalid name' % col
        self.array[self.row.index(row), self.col.index(col)] = val


class MathArrayFunc(object):
    """A class of basic array functions some LAPACK based"""

    __doc__ = '''PySCeS array functions - used by Stoich'''

    array_kind = {'i': 0, 'l': 0, 'f': 0, 'd': 0, 'F': 1, 'D': 1}
    array_precision = {'i': 1, 'l': 1, 'f': 0, 'd': 1, 'F': 0, 'D': 1}
    array_type = [['f', 'd'], ['F', 'D']]
    LinAlgError = 'LinearAlgebraError'

    def commonType(self, *arrays):
        """
        commonType(\*arrays)

        Numeric detect and set array precision (will be replaced with new scipy.core compatible code when ready)

        Arguments:

        \*arrays: input arrays

        """
        kind = 0
        #    precision = 0
        #   force higher precision in lite version
        precision = 1
        for a in arrays:
            t = a.dtype.char
            kind = max(kind, self.array_kind[t])
            precision = max(precision, self.array_precision[t])

        return self.array_type[kind][precision]

    def castCopyAndTranspose(self, type, *arrays):
        """
        castCopyAndTranspose(type, \*arrays)

        Cast numeric arrays to required type and transpose

        Arguments:

        type: the required type to cast to
        \*arrays: the arrays to be processed

        """
        cast_arrays = ()
        for a in arrays:
            if a.typecode() == type:
                cast_arrays = cast_arrays + (copy.copy(scipy.transpose(a)),)
            else:
                cast_arrays = cast_arrays + (
                    copy.copy(scipy.transpose(a).astype(type)),
                )
        if len(cast_arrays) == 1:
            return cast_arrays[0]
        else:
            return cast_arrays

    def assertRank2(self, *arrays):
        """
        assertRank2(\*arrays)

        Check that we are using a 2D array

        Arguments:

        \*arrays: input array(s)

        """
        for a in arrays:
            if len(a.shape) != 2:
                raise LinAlgError('Array must be two-dimensional')

    def SwapCol(self, res_a, r1, r2):
        """
        SwapCol(res_a,r1,r2)

        Swap two columns using BLAS swap, arrays can be (or are upcast to) type double (d) or double complex (D).
        Returns the colswapped array

        Arguments:

        res_a: the input array
        r1: the first column to be swapped
        r2: the second column to be swapped

        """
        if (
            self.array_kind[self.commonType(res_a)] == 1
        ):  # brett 20041226 added complex support to swap
            res_a = res_a.astype('D')
            return self.SwapColz(res_a, r1, r2)
        else:
            res_a = res_a.astype('d')
            return self.SwapCold(res_a, r1, r2)

    def SwapRow(self, res_a, r1, r2):
        """
        SwapRow(res_a,r1,r2)

        Swaps two rows using BLAS swap, arrays can be (or are upcast to) type double (d) or double complex (D).
        Returns the rowswapped array.

        Arguments:

        res_a: the input array
        r1: the first row index to be swapped
        r2:  the second row index to be swapped

        """
        if (
            self.array_kind[self.commonType(res_a)] == 1
        ):  # brett 20041226 added complex support to swap
            res_a = res_a.astype('D')
            return self.SwapRowz(res_a, r1, r2)
        else:
            res_a = res_a.astype('d')
            return self.SwapRowd(res_a, r1, r2)

    def SwapElem(self, res_a, r1, r2):
        """
        SwapElem(res_a,r1,r2)

        Swaps two elements in a 1D vector

        Arguments:

        res_a: the input vector
        r1: index 1
        r2: index 2

        """
        res_a[r1], res_a[r2] = res_a[r2], res_a[r1]
        return res_a

    def SwapCold(self, res_a, c1, c2):
        """
        SwapCold(res_a,c1,c2)

        Swaps two double (d) columns in an array using BLAS DSWAP. Returns the colswapped array.

        Arguments:

        res_a: input array
        c1: column index 1
        c2: column index 2

        """
        res_a = res_a.astype('d')
        res_a[:, c1], res_a[:, c2] = myfblas.dswap(res_a[:, c1], res_a[:, c2])
        return res_a

    def SwapRowd(self, res_a, r1, r2):
        """
        SwapRowd(res_a,c1,c2)

        Swaps two double (d) rows in an array using BLAS DSWAP. Returns the rowswapped array.

        Arguments:

        res_a: input array
        c1: row index 1
        c2: row index 2

        """
        res_a = res_a.astype('d')
        res_a[r1, :], res_a[r2, :] = myfblas.dswap(res_a[r1, :], res_a[r2, :])
        return res_a

    def SwapColz(self, res_a, c1, c2):
        """
        SwapColz(res_a,c1,c2)

        Swaps two double complex (D) columns in an array using BLAS ZSWAP. Returns the colswapped array.

        Arguments:

        res_a: input array
        c1: column index 1
        c2: column index 2

        """
        res_a = res_a.astype('D')
        res_a[:, c1], res_a[:, c2] = myfblas.zswap(res_a[:, c1], res_a[:, c2])
        return res_a

    def SwapRowz(self, res_a, r1, r2):
        """
        SwapRowz(res_a,c1,c2)

        Swaps two double complex (D) rows in an array using BLAS ZSWAP. Returns the rowswapped array.

        Arguments:

        res_a: input array
        c1: row index 1
        c2: row index 2

        """
        res_a = res_a.astype('D')
        res_a[r1, :], res_a[r2, :] = myfblas.zswap(res_a[r1, :], res_a[r2, :])
        return res_a

    def MatrixFloatFix(self, mat, val=1.0e-15):
        """
        MatrixFloatFix(mat,val=1.e-15)

        Clean an array removing any floating point artifacts defined as being smaller than a specified value.
        Processes an array inplace

        Arguments:

        mat: the input 2D array
        val [default=1.e-15]: the threshold value (effective zero)

        """
        zero_vals = abs(mat) < val

        for x in range(mat.shape[0]):
            for y in range(mat.shape[1]):
                if zero_vals[x, y]:
                    mat[x, y] = 0.0
                # if abs(mat[x,y]) != 0.0 and abs(mat[x,y]) < val:
                ##  mat[x,y] = round(mat[x,y])
                # if abs(mat[x,y]) < val:
                # mat[x,y] = round(mat[x,y])
                ##  mat[x,y] = 0.0
        del zero_vals

    def MatrixValueCompare(self, matrix):
        """
        MatrixValueCompare(matrix)

        Finds the largest/smallest abs(value) > 0.0 in a matrix.
        Returns a tuple containing (smallest,largest) values

        Arguments:

        matrix: the input 2D array

        """
        val_B = 0.0
        val_S = 1.0e30
        for x in range(matrix.shape[0]):
            for y in range(matrix.shape[1]):
                if abs(matrix[x, y]) > abs(val_B) and abs(matrix[x, y]) != 0.0:
                    val_B = matrix[x, y]
                if abs(matrix[x, y]) < abs(val_S) and abs(matrix[x, y]) != 0.0:
                    val_S = matrix[x, y]
        return (val_S, val_B)


class Stoich(MathArrayFunc):
    '''PySCeS stoichiometric analysis class: initialized with a stoichiometric matrix N (input)'''

    __stoichdiagmode__ = 0
    __version__ = __version__
    __TimeFormat = "%H:%M:%S"
    USE_QR = False
    info_moiety_conserve = False

    def __init__(self, input):
        """Initialize class variables"""

        self.nmatrix = input
        row, col = self.nmatrix.shape
        self.nmatrix_col = tuple(scipy.array(range(col)))
        self.nmatrix_row = tuple(scipy.array(range(row)))

        # Create a machine specific instance
        from scipy import MachAr

        mach_spec = MachAr()

        self.stoichiometric_analysis_fp_zero = mach_spec.eps * 2.0e4
        self.stoichiometric_analysis_lu_precision = self.stoichiometric_analysis_fp_zero
        self.stoichiometric_analysis_gj_precision = (
            self.stoichiometric_analysis_lu_precision * 10.0
        )

        self.species = None
        self.reactions = None

    def AnalyseK(self):
        """
        AnalyseK()

        Evaluate the stoichiometric matrix and calculate the nullspace using LU decomposition and backsubstitution .
        Generates the MCA K and Ko arrays and associated row and column vectors

        Arguments:
        None

        """
        print('Calculating K matrix .',)

        self.info_flux_conserve = 0  # added 20020416 for conservation detection

        if self.__stoichdiagmode__:
            print('\nKMATRIX\nGetUpperMatrix: ' + time.strftime(self.__TimeFormat))

        p, u, row_vector, column_vector = self.GetUpperMatrix(self.nmatrix)
        # p,u,row_vector,column_vector = self.GetUpperMatrixUsingQR(self.nmatrix)
        if self.__stoichdiagmode__:
            print('\nKMATRIX\nScalePivots: ' + time.strftime(self.__TimeFormat))

        unipiv_a = self.ScalePivots(u)
        if self.__stoichdiagmode__:
            print('\nKMATRIX\nBackSubstitution: ' + time.strftime(self.__TimeFormat))

        R_a, row_vector, column_vector = self.BackSubstitution(
            unipiv_a, row_vector, column_vector
        )
        if self.__stoichdiagmode__:
            print('\nKMATRIX\nK_split_R: ' + time.strftime(self.__TimeFormat))

        (
            r_ipart,
            r_fpart,
            row_vector,
            column_vector,
            nullspace,
            r_fcolumns,
            self.info_flux_conserve,
        ) = self.K_split_R(R_a, row_vector, column_vector)
        # Don't need anymore ... I think - brett 20051013
        # try:
        # self.MatrixFloatFix(nullspace,val=self.stoichiometric_analysis_lu_precision)
        # except Exception as e:
        # if self.__stoichdiagmode__:
        # print('Ignored (no K)',e) # brett 20050801

        if self.__stoichdiagmode__:
            print('\nKMATRIX\n' + time.strftime(self.__TimeFormat))

        self.kmatrix = nullspace
        self.kmatrix_row = tuple(row_vector)
        self.kmatrix_col = tuple(column_vector)

        self.kzeromatrix = r_fpart
        self.kzeromatrix_row = tuple(r_fcolumns)
        self.kzeromatrix_col = tuple(column_vector)
        print(' done.')

    def AnalyseL(self):
        """
        AnalyseL()

        Evaluate the stoichiometric matrix and calculate the left nullspace using LU factorization and backsubstitution.
        Generates the MCA L, Lo, Nr and Conservation matrix and associated row and column vectors

        Arguments:
        None

        """
        print('Calculating L matrix .',)

        a = scipy.transpose(self.nmatrix)
        self.info_moiety_conserve = False  # added 20020416 for conservation detection

        if self.__stoichdiagmode__:
            print('\nLMATRIX\nGetUpperMatrix: ' + time.strftime(self.__TimeFormat))

        if not self.USE_QR:
            p, u, row_vector, column_vector = self.GetUpperMatrix(a)
        else:
            p, u, row_vector, column_vector = self.GetUpperMatrixUsingQR(a)
        if self.__stoichdiagmode__:
            print('\nLMATRIX\nScalePivots: ' + time.strftime(self.__TimeFormat))
        unipiv_a = self.ScalePivots(u)
        if self.__stoichdiagmode__:
            print('\nLMATRIX\nBackSubstitution: ' + time.strftime(self.__TimeFormat))

        R_a, row_vector, column_vector = self.BackSubstitution(
            unipiv_a, row_vector, column_vector
        )
        if self.__stoichdiagmode__:
            print('\nLMATRIX\nK_split_R: ' + time.strftime(self.__TimeFormat))

        (
            r_ipart,
            consmatrix,
            cons_row_vector,
            cons_col_vector,
            lmatrix,
            lmatrix_row_vector,
            lmatrix_col_vector,
            lomatrix,
            lomatrix_row_vector,
            lomatrix_col_vector,
            nrmatrix,
            Nred_vector,
            Nred_vector_col,
            self.info_moiety_conserve,
        ) = self.L_split_R(a, R_a, row_vector, column_vector)
        # Don't need anymore ... i think - brett 20051013
        # try:
        # self.MatrixFloatFix(consmatrix,val=self.stoichiometric_analysis_lu_precision)
        # except Exception as e:
        # if self.__stoichdiagmode__:
        # print('Ignored (no L)',e) # brett 20050801

        if self.__stoichdiagmode__:
            print('\nLMATRIX\n' + time.strftime(self.__TimeFormat))

        self.lmatrix = lmatrix
        self.lmatrix_row = tuple(lmatrix_row_vector)
        self.lmatrix_col = tuple(lmatrix_col_vector)

        self.lzeromatrix = lomatrix
        self.lzeromatrix_row = tuple(lomatrix_row_vector)
        self.lzeromatrix_col = tuple(lomatrix_col_vector)

        self.conservation_matrix = consmatrix
        self.conservation_matrix_row = tuple(cons_row_vector)
        self.conservation_matrix_col = tuple(cons_col_vector)

        self.nrmatrix = nrmatrix
        self.nrmatrix_row = tuple(Nred_vector)
        self.nrmatrix_col = tuple(scipy.copy(self.nmatrix_col))
        print(' done.')

    def PivotSort(self, a, row_vector, column_vector):
        """
        PivotSort(a,row_vector,column_vector)

        This is a sorting routine that accepts a matrix and row/colum vectors
        and then sorts them so that: there are no zero rows (by swapping with first
        non-zero row) The abs(largest) pivots are moved onto the diagonal to maintain
        numerical stability. Row and column swaps are recorded in the tracking vectors.

        Arguments:

        a: the input array
        row_vector: row tracking vector
        column_vector: column tracking vector

        """
        t = self.commonType(a)
        row, col = a.shape

        for z in range(0, min(row, col)):
            if abs(a[z, z]) < self.stoichiometric_analysis_lu_precision:
                maxV = self.stoichiometric_analysis_lu_precision
                maxP = (None, None)
                zeroP = (None, None)

                mList = []
                for x in range(z, min(row, col)):
                    mList.append((x, max(abs(a[x, x:]))))
                mVal = 0.0
                mRow = None
                # print(mList)
                for el in mList:
                    if el[1] > mVal:
                        mVal = el[1]
                        mRow = el[0]
                if self.__stoichdiagmode__:
                    print('\tmVal', mVal, mRow)

                if mRow != None:
                    for y in range(z, col):
                        if (
                            abs(a[mRow, y]) > abs(maxV)
                            and abs(a[mRow, y])
                            > self.stoichiometric_analysis_lu_precision
                        ):
                            maxV = a[mRow, y]
                            maxP = (mRow, y)

                if zeroP != (None, None) and zeroP != (z, z) and mRow != None:
                    if self.__stoichdiagmode__:
                        print('  Swapping: ', (z, z), (zeroP[0], zeroP[1]), 1.0)
                    a = self.SwapRowd(a, z, zeroP[0])
                    a = self.SwapCold(a, z, zeroP[1])
                    row_vector = self.SwapElem(row_vector, z, zeroP[0])
                    column_vector = self.SwapElem(column_vector, z, zeroP[1])
                elif (
                    maxP != (None, None)
                    and maxP != (min(row, col), min(row, col))
                    and mRow != None
                ):
                    if self.__stoichdiagmode__:
                        print('  Swapping: ', (z, z), (maxP[0], maxP[1]), a[z, z], maxV)
                    a = self.SwapRowd(a, z, maxP[0])
                    a = self.SwapCold(a, z, maxP[1])
                    row_vector = self.SwapElem(row_vector, z, maxP[0])
                    column_vector = self.SwapElem(column_vector, z, maxP[1])

        return (a, row_vector, column_vector)

    def PivotSort_initial(self, a, row_vector, column_vector):
        """
        PivotSort_initial(a,row_vector,column_vector)

        This is a sorting routine that accepts a matrix and row/colum vectors
        and then sorts them so that: the abs(largest) pivots are moved onto the diagonal to maintain
        numerical stability i.e. the matrix diagonal is in descending max(abs(value)).
        Row and column swaps are recorded in the tracking vectors.

        Arguments:

        a: the input array
        row_vector: row tracking vector
        column_vector: column tracking vector

        """

        # SAME AS THE ABOVE JUST DOES ALL VALUES NOT ONLY NON_ZERO ONES
        # brett 200500802

        t = self.commonType(a)
        row, col = a.shape

        for z in range(0, min(row, col)):
            if z == 0 or abs(a[z, z]) < abs(a[z - 1, z - 1]):
                maxV = self.stoichiometric_analysis_lu_precision
                maxP = (None, None)
                zeroP = (None, None)
                mList = []
                for x in range(z, min(row, col)):
                    mList.append((x, max(abs(a[x, x:]))))
                mVal = 0.0
                mRow = None
                for el in mList:
                    if el[1] > mVal:
                        mVal = el[1]
                        mRow = el[0]
                if self.__stoichdiagmode__:
                    print('\tmVal', mVal, mRow)

                if mRow != None:
                    for y in range(z, col):
                        if (
                            abs(a[x, y]) > abs(maxV)
                            and abs(a[x, y]) > self.stoichiometric_analysis_lu_precision
                        ):
                            maxV = a[x, y]
                            maxP = (x, y)

                if (
                    maxP != (None, None)
                    and maxP != (min(row, col), min(row, col))
                    and maxP != (z, z)
                ):
                    if self.__stoichdiagmode__:
                        print('  Swapping: ', (z, z), (maxP[0], maxP[1]), a[z, z], maxV)
                    a = self.SwapRowd(a, z, maxP[0])
                    a = self.SwapCold(a, z, maxP[1])
                    row_vector = self.SwapElem(row_vector, z, maxP[0])
                    column_vector = self.SwapElem(column_vector, z, maxP[1])

        return (a, row_vector, column_vector)

    def PLUfactorize(self, a_in):
        """
        PLUfactorize(a_in)

        Performs an LU factorization using LAPACK D/ZGetrf. Now optimized for FLAPACK interface.
        Returns LU - combined factorization, IP - rowswap information and info - Getrf error control.

        Arguments:

        a_in: the matrix to be factorized

        """
        print('.',)

        self.assertRank2(a_in)
        t = self.commonType(a_in)
        if a_in.dtype.char == 'D':
            # print('Complex matrix ' + a_in.typecode())
            ##  a = copy.copy(scipy.transpose(a_in))
            # brett 201106 flapack optimize
            a = a_in.copy()
        elif a_in.dtype.char == 'd':
            # print('Float matrix ' + a_in.typecode())
            ##  a = copy.copy(scipy.transpose(a_in))
            # brett 201106 flapack optimize
            a = a_in.copy()
        else:
            # print('Other matrix casting to double, was: ' + a_in.typecode())
            ##  a = copy.copy(scipy.transpose(a_in).astype('d'))
            # brett 201106 flapack optimize
            a = a_in.copy().astype('d')
        Using_FLAPACK = 1
        # brett 20110620 - clapack support is being removed from scipy going exclusively flapack now
        # brett 20041226 - protecting ourselves against flapack
        # try:
        # scipy.linalg.clapack.empty_module()
        ##  Using_FLAPACK = 1
        ##  print("Using FLAPACK")
        # except:
        ##  print("Using CLAPACK")

        if Using_FLAPACK == 1:
            try:
                if self.array_kind[t] == 1:
                    getrf = myflapack.zgetrf  # scipy flapack 20041226
                else:
                    getrf = myflapack.dgetrf  # scipy flapack 20041226
            except Exception as e:
                print("FLAPACK error", e)
        # else:
        # try:
        # if self.array_kind[t] == 1:
        # getrf = scipy.linalg.clapack.zgetrf #scipy clapack (ATLAS) 20030605
        # else:
        # getrf = scipy.linalg.clapack.dgetrf #scipy clapack (ATLAS) 20030605
        # except Exception as e:
        ##  print("CLAPACK error", e)

        if Using_FLAPACK == 1:
            # results = getrf(scipy.transpose(a)) # brett 20041226
            results = getrf(a)  # brett 201106
        # else:
        # This is a $%^&*( ... suddenly with latest cvs f2py getrf only accepts arrays
        # not vectors so this song and dance is necessary to fix this 'behaviour'
        # as idiotic as it seems, we pad the vector with a zero row (no difference numerically)
        # run it through getrf and then strip it afterwards !@#$%^&*() - brett 20050707
        # if a.shape[0] == 1:
        ##  tarr = scipy.zeros((a.shape[0]+1,a.shape[1]),'d')
        ##  tarr[0] = a[0]

        # results = getrf(tarr) #scipy cblas (ATLAS) 20030506 -- this is normal

        ##  results = list(results)
        ##  results[0] = scipy.array([results[0][0]])
        ##  results[1] = scipy.array([results[1][0]],'i')
        ##  results[2] = results[2]
        ##  results = tuple(results)
        # else:
        # results = getrf(a) #scipy cblas (ATLAS) 20030506 -- this is normal

        results = list(results)

        if results[2] < 0:
            print('Argument ', results['info'], ' had an illegal value')
            raise LinAlgError
        elif results[2] > 0:
            pass

        if Using_FLAPACK == 1:
            result = results[0]  # brett 20041226
        # else:
        # result = scipy.transpose(results[0]) # -- this is normal

        badlist = []
        for x in range(result.shape[0]):
            for y in range(result.shape[1]):
                if (
                    abs(result[x, y]) != 0.0
                    and abs(result[x, y]) < self.stoichiometric_analysis_lu_precision
                ):
                    result[x, y] = 0.0  # gets rid of the -0.0 irritation
                    if len(badlist) == 0 and (x, y) == (
                        x,
                        x,
                    ):  # catch 1st float on a pivot
                        badlist.append(x)
        if len(badlist) != 0:
            results[2] = (
                badlist[0] - 1
            )  # 20040423 if float was on a pivot - refactorize

        return (result, results[1], results[2])  # scipy cblas (ATLAS?) 20030506

    # def PLUfactorizeOLD(self,a_in):
    # """
    # PLUfactorize(a_in) OLD PRE SCIPY 0.10 uses clapack

    # Performs an LU factorization using LAPACK D/ZGetrf.
    # Returns LU - combined factorization, IP - rowswap information and info - Getrf error control.

    # Arguments:

    # a_in: the matrix to be factorized

    # """
    # print('.',)

    # self.assertRank2(a_in)
    ##  t = self.commonType(a_in)
    # if a_in.dtype.char == 'D':
    # print 'Complex matrix ' + a_in.typecode()
    ##  a = copy.copy(scipy.transpose(a_in))
    # elif a_in.dtype.char == 'd':
    # print 'Float matrix ' + a_in.typecode()
    ##  a = copy.copy(scipy.transpose(a_in))
    # else:
    # print 'Other matrix casting to double, was: ' + a_in.typecode()
    ##  a = copy.copy(scipy.transpose(a_in).astype('d'))
    ##  Using_FLAPACK = 0
    # brett 20041226 - protecting ourselves against flapack
    # try:
    # scipy.linalg.clapack.empty_module()
    ##  Using_FLAPACK = 1
    # print "Using FLAPACK"
    # except:
    # print "Using CLAPACK"

    # if Using_FLAPACK == 1:
    # try:
    # if self.array_kind[t] == 1:
    # getrf = myflapack.zgetrf #scipy flapack 20041226
    # else:
    # getrf = myflapack.dgetrf #scipy flapack 20041226
    # except Exception as e:
    # print "FLAPACK error", e
    # else:
    # try:
    # if self.array_kind[t] == 1:
    # getrf = scipy.linalg.clapack.zgetrf #scipy clapack (ATLAS) 20030605
    # else:
    # getrf = scipy.linalg.clapack.dgetrf #scipy clapack (ATLAS) 20030605
    # except Exception as e:
    # print "CLAPACK error", e

    # if Using_FLAPACK == 1:
    # results = getrf(scipy.transpose(a)) # brett 20041226
    # else:
    # This is a $%^&*( ... suddenly with latest cvs f2py getrf only accepts arrays
    # not vectors so this song and dance is necessary to fix this 'behaviour'
    # as idiotic as it seems, we pad the vector with a zero row (no difference numerically)
    # run it through getrf and then strip it afterwards !@#$%^&*() - brett 20050707
    # if a.shape[0] == 1:
    ##  tarr = scipy.zeros((a.shape[0]+1,a.shape[1]),'d')
    ##  tarr[0] = a[0]

    # results = getrf(tarr) #scipy cblas (ATLAS) 20030506 -- this is normal

    ##  results = list(results)
    ##  results[0] = scipy.array([results[0][0]])
    ##  results[1] = scipy.array([results[1][0]],'i')
    ##  results[2] = results[2]
    ##  results = tuple(results)
    # else:
    # results = getrf(a) #scipy cblas (ATLAS) 20030506 -- this is normal

    ##  results = list(results)

    # if results[2] < 0:
    ##  print('Argument ', results['info'], ' had an illegal value')
    ##  raise LinAlgError
    # elif results[2] > 0:
    # pass

    # if Using_FLAPACK == 1:
    # result = results[0] # brett 20041226
    # else:
    # result = scipy.transpose(results[0]) # -- this is normal

    ##  badlist = []
    # for x in range(result.shape[0]):
    # for y in range(result.shape[1]):
    # if abs(result[x,y]) != 0.0 and abs(result[x,y]) < self.stoichiometric_analysis_lu_precision:
    # result[x,y] = 0.0        # gets rid of the -0.0 irritation
    # if len(badlist) == 0 and (x,y) == (x,x): # catch 1st float on a pivot
    # badlist.append(x)
    # if len(badlist) != 0:
    # results[2] = badlist[0]-1 # 20040423 if float was on a pivot - refactorize

    # return(result,results[1],results[2]) #scipy cblas (ATLAS?) 20030506

    def SplitLU(self, plu, row, col, t=None):
        """
        SplitLU(plu,row,col,t)

        PLU takes the combined LU factorization computed by PLUfactorize and extracts the upper matrix.
        Returns U.

        Arguments:

        plu: LU factorization
        row: row tracking vector
        col: column tracking vector
        t [default=None)]: typecode argument (currently not used)

        """
        print('.',)

        if self.__stoichdiagmode__:
            print('\n', row, col)
        for cl in range(0, min(row, col)):
            plu[cl + 1 :, cl] = 0.0
        return plu

    def GetUpperMatrix(self, a):
        """
        GetUpperMatrix(a)

        Core analysis algorithm; an input is preconditioned using PivotSort_initial and then cycles of PLUfactorize and
        PivotSort are run until the factorization is completed. During this process the matrix is reordered by
        column swaps which emulates a full pivoting LU factorization. Returns the pivot matrix P, upper factorization U
        as well as the row/col tracking vectors.

        Arguments:

        a: a stoichiometric matrix

        """
        print('.',)

        t = self.commonType(a)
        row, col = a.shape

        # this is a test brett 20050802
        row_vector = scipy.array((range(row)))
        column_vector = scipy.array((range(col)))
        a, row_vector, column_vector = self.PivotSort_initial(
            a, row_vector, column_vector
        )

        # 18/09/2000 PLUfactorize included in self.GetUpperMatrix
        a, ip, info = self.PLUfactorize(a)

        # get U from getrf's PLU
        if self.__stoichdiagmode__:
            print(info)
        upper_out = self.SplitLU(a, row, col, t)

        p_out = scipy.identity(row)
        for x in range(0, min(row, col)):
            if x + 1 != ip[x]:
                p_out = self.SwapRowd(p_out, x, (ip[x] - 1))
                row_vector = self.SwapElem(row_vector, x, (ip[x] - 1))

        '''31/08/2000 Added to try to make sure that pivots exist only on the upper left side of the matrix
        max(abs(upper_out[x,:])) used instead of abs(max(upper_out[x,:])) otherwise for rows where
        all elements < 0 zero is maximum, this way the max of positive values is used'''

        upper_out, row_vector, column_vector = self.PivotSort(
            upper_out, row_vector, column_vector
        )

        '''20/09/2000 This bit sorts out the echelon matrix by running (if needed) cycles of
        self.PLUfactorize, self.GetUpperMatrix, and pivsort until only a staircase matrix remains. It uses both the error
        generated by self.PLUfactorize(info) and go_flag to control itself'''

        if info > 0:
            # Here we check to see if the error is not in the last or reduced-last column if it is - exit
            #        go_flag = 'yes' # 2001/04/26 changed for Python21 and future compatibility
            go_flag = 1
            for x in range(0, min(row, col)):
                if abs(upper_out[x, x]) > self.stoichiometric_analysis_lu_precision:
                    pos_holder = x
            if pos_holder + 1 < info and pos_holder + 1 < row:
                if (
                    max(abs(upper_out[pos_holder + 1, :]))
                    < self.stoichiometric_analysis_lu_precision
                ):
                    #                go_flag = 'no' # 2001/04/26 changed for Python21 and future compatibility
                    go_flag = 0
            elif pos_holder + 1 == info and pos_holder + 1 == min(row, col):
                #            go_flag = 'no' # 2001/04/26 changed for Python21 and future compatibility
                go_flag = 0
            #        while info > 0 and go_flag == 'yes': # 2001/04/26 changed for Python21 and future compatibility
            while info > 0 and go_flag == 1:
                # sort
                # upper_out,row_vector,column_vector = pivot_sort2k5(upper_out,row_vector,column_vector)
                upper_out, row_vector, column_vector = self.PivotSort(
                    upper_out, row_vector, column_vector
                )
                # PLUfactorize
                if self.__stoichdiagmode__:
                    print(
                        ' Echelon: ' + time.strftime(self.__TimeFormat), '(', info, ')'
                    )
                upper_out, ip, info = self.PLUfactorize(upper_out)

                # get U from getrf's PLU
                if self.__stoichdiagmode__:
                    print(info)
                upper_out = self.SplitLU(upper_out, row, col, t)

                # realign row vector and permutation matrix if necessary
                for x in range(0, min(row, col)):
                    if x + 1 != ip[x]:
                        p_out = self.SwapRowd(p_out, x, (ip[x] - 1))
                        row_vector = self.SwapElem(row_vector, x, (ip[x] - 1))
                # test if we can exit -- same criteria as before
                for x in range(0, min(row, col)):
                    if abs(upper_out[x, x]) > self.stoichiometric_analysis_lu_precision:
                        pos_holder = x
                if pos_holder + 1 < info and pos_holder + 1 < row:
                    if (
                        max(abs(upper_out[pos_holder + 1, :]))
                        < self.stoichiometric_analysis_lu_precision
                    ):
                        #                    go_flag = 'no' # 2001/04/26 changed for Python21 and future compatibility
                        go_flag = 0
                elif pos_holder + 1 == info and pos_holder + 1 == min(row, col):
                    #                go_flag = 'no' # 2001/04/26 changed for Python21 and future compatibility
                    go_flag = 0

        '''This bit will get rid of any zero rows so that we will hopefully only have to work with a
        reduced matrix after this, for completeness the row_vector will also be sliced'''

        for x in range(0, min(row, col)):
            if abs(upper_out[x, x]) > self.stoichiometric_analysis_lu_precision:
                pos_holder = x

        upper_out_r = scipy.zeros((pos_holder + 1, col)).astype(t)
        upper_out_r = upper_out[: pos_holder + 1, :]
        row_vector_r = row_vector[: pos_holder + 1]

        return (p_out, upper_out_r, row_vector_r, column_vector)

    def GetUpperMatrixUsingQR(self, a):
        """
        GetUpperMatrix(a)

        Core analysis algorithm; an input is preconditioned using PivotSort_initial and then cycles of PLUfactorize and
        PivotSort are run until the factorization is completed. During this process the matrix is reordered by
        column swaps which emulates a full pivoting LU factorization. Returns the pivot matrix P, upper factorization U
        as well as the row/col tracking vectors.

        Arguments:

        a: a stoichiometric matrix

        """
        print('.',)

        t = self.commonType(a)
        row, col = a.shape

        # this is a test brett 20050802
        row_vector = scipy.array((range(row)))
        column_vector = scipy.array((range(col)))
        ##  a,row_vector,column_vector = self.PivotSort(a,row_vector,column_vector)
        a, row_vector, column_vector = self.PivotSort_initial(
            a, row_vector, column_vector
        )

        Q, upper_out = scipy.linalg.qr(a)
        self.MatrixFloatFix(
            upper_out, val=self.stoichiometric_analysis_lu_precision * 10.0
        )

        '''This bit will get rid of any zero rows so that we will hopefully only have to work with a
        reduced matrix after this, for completeness the row_vector will also be sliced'''

        for x in range(0, min(row, col)):
            if abs(upper_out[x, x]) > self.stoichiometric_analysis_lu_precision:
                pos_holder = x

        upper_out_r = scipy.zeros((pos_holder + 1, col)).astype(t)
        p_out = scipy.diag(upper_out.shape[1] * [1.0])
        upper_out_r = upper_out[: pos_holder + 1, :]
        row_vector_r = row_vector[: pos_holder + 1]

        return (p_out, upper_out_r, row_vector_r, column_vector)

    def ScalePivots(self, a_one):
        """
        ScalePivots(a_one)

        Given an upper triangular matrix U, this method scales the diagonal (pivot values) to one.

        Arguments:

        a_one: an upper triangular matrix U

        """
        print('.',)

        t = self.commonType(a_one)
        row, col = a_one.shape

        '''13/09/2000 We now assume that the matrix has the correct shape ie. the pivots are in a
        perfect staircase'''

        for x in range(0, row):
            if (
                abs(a_one[x, x]) == 0.0
            ):  # ths is to prevent NAN's when FixFloat zeros a supersmall pivot
                pass
            elif abs(a_one[x, x]) != 1.0:
                a_one[x, :] = a_one[x, :] / abs(a_one[x, x])
            if a_one[x, x] < 0.0:
                a_one[x, :] = -(a_one[x, :])
        return a_one

    def BackSubstitution(self, res_a, row_vector, column_vector):
        """
        BackSubstitution(res_a,row_vector,column_vector)

        Jordan reduction of a scaled upper triangular matrix. The returned array is now in the form [I R] and can
        be used for nullspace determination. Modified row and column tracking vetors are also returned.

        Arguments:

        res_a: unitary pivot upper triangular matrix
        row_vector: row tracking vector
        column_vector: column tracking vector

        """
        print('.',)

        t = self.commonType(res_a)
        row, col = res_a.shape

        '''13/09/2000 removed the copy thing, and eliminated the divnumb variable. Because we
        are now working with a row reduced matrix upward elimination only works on the pivot cols'''

        # old back substitution circa 2000! brett - 20050801
        ##  bigF = 0
        # if max(res_a.shape) > 500:
        ##  bigF = 1
        # for y in range (min(row,col)-1,-1,-1):
        # if bigF and self.__stoichdiagmode__:
        # print('.',)
        # for x in range (min(row,col)-2,-1,-1):
        ##  z = x + 1
        # while z < min(row,col):
        # if abs(res_a[x,y]) > self.stoichiometric_analysis_gj_precision and abs(res_a[z,y]) > self.stoichiometric_analysis_gj_precision:
        ##  res_a[x,:] = res_a[x,:] - res_a[z,:]*(res_a[x,y]/res_a[z,y])
        ##  z = z + 1
        # continue
        # else:
        ##  z = z + 1

        # new back substitution
        # right looking algorithm that uses array slicing speeds up as it moves down the pivots
        # this algorithm is at least 3X as fast as the old one for a large system - brett 20050805
        bigF = 0
        if max(res_a.shape) > 500:
            bigF = 1
        for x in range(min(row, col)):
            if bigF and self.__stoichdiagmode__:
                print('.',)
            for y in range(x + 1, min(row, col)):
                if abs(res_a[x, y]) > self.stoichiometric_analysis_lu_precision:
                    res_a[x, y:] = res_a[x, y:] - (res_a[x, y] * res_a[y, y:])

        # $%^&* wtf is this for??? brett - 20050801 (besides cleaning fp wierdness ... nothing ;-)
        self.MatrixFloatFix(res_a, val=self.stoichiometric_analysis_gj_precision)
        res_a, row_vector, column_vector = self.PivotSort(
            res_a, row_vector, column_vector
        )

        return (res_a, row_vector, column_vector)

    def K_split_R(self, R_a, row_vector, column_vector):
        """
        K_split_R(R_a,row_vector,column_vector)

        Using the R factorized form of the stoichiometric matrix we now form the K and Ko matrices. Returns
        the r_ipart,Komatrix,Krow,Kcolumn,Kmatrix,Korow,info

        Arguments:

        R_a: the Gauss-Jordan reduced stoichiometric matrix
        row_vector: row tracking vector
        column_vector: column tracking vector

        """
        print('.',)

        t = self.commonType(R_a)
        row, col = R_a.shape

        '''14/09/2000 Seeing as we now should have a perfect staircase in a reduced matrix we do
        not have to search for the last pivot it will exist at min(row,col)-1 so the pos_holder
        finding code has been removed (actually moved into self.GetUpperMatrix)'''

        pos_holder = min(row, col) - 1

        '''This bit extracts the identity part from R (future note this could be replaced by an I matrix formed by min(row,col))'''

        r_ipart = scipy.zeros((pos_holder + 1, pos_holder + 1)).astype(t)
        r_ipart = R_a[: pos_holder + 1, : pos_holder + 1]

        row_i, col_i = r_ipart.shape

        '''If there are free variables, then this bit will extract them and form the row/col vectors'''

        empty_rf = 0
        K_switch = 0  # added 20020416 class attribute for conservation detection 0 = none, 1 = exists

        if col - col_i > self.stoichiometric_analysis_fp_zero:
            r_fpart = scipy.zeros((pos_holder + 1, col - (pos_holder + 1))).astype(t)
            r_fpart = R_a[: pos_holder + 1, pos_holder + 1 :]

            row_vector_dependent = column_vector[: pos_holder + 1]
            row_vector_independent = column_vector[pos_holder + 1 :]

            # row_vector = scipy.concatenate((row_vector_independent,row_vector_dependent),1)
            row_vector = scipy.hstack((row_vector_independent, row_vector_dependent))

            column_vector = column_vector[pos_holder + 1 :]
            K_switch = 1
        else:
            print('no flux conservation')
            r_fpart = 'no flux conservation'
            empty_rf = 1
            K_switch = 0

        #    if r_fpart == 'F is empty, R = I': # 2001/04/26 changed for Python21 compatibility
        if empty_rf == 1:
            return (
                r_ipart,
                r_ipart,
                column_vector,
                column_vector,
                r_ipart,
                column_vector,
                K_switch,
            )
        else:
            row, col = r_fpart.shape
            id = scipy.identity(col).astype(t)
            nullspace = scipy.concatenate((id, -r_fpart), 0).astype(t)

        # brett 05/11/2002 changed r_fpart to -r_fpart
        return (
            r_ipart,
            -r_fpart,
            row_vector,
            column_vector,
            nullspace,
            row_vector_dependent,
            K_switch,
        )

    def L_split_R(self, Nfull, R_a, row_vector, column_vector):
        """
        L_split_R(Nfull,R_a,row_vector,column_vector)

        Takes the Gauss-Jordan factorized N^T and extract the L, Lo, conservation (I -Lo) and reduced stoichiometric matrices. Returns: lmatrix_col_vector, lomatrix, lomatrix_row, lomatrix_co, nrmatrix, Nred_vector_row, Nred_vector_col, info

        Arguments:

        Nfull: the original stoichiometric matrix N
        R_a: gauss-jordan factorized form of N^T
        row_vector: row tracking vector
        column_vector: column tracking vector

        """
        print('.',)

        # print '\nR_a'
        # print str(R_a)
        t = self.commonType(R_a)
        row, col = R_a.shape

        '''14/09/2000 Seeing as we now should have a perfect staircase in a reduced matrix we do
        not have to search for the last pivot it will exist at min(row,col)-1 so the pos_holder
        finding code has been removed (actually moved into self.GetUpperMatrix)'''

        pos_holder = min(row, col) - 1

        '''Here we extract the identity matrix from R (future note this could be replaced by an I matrix formed by min(row,col))'''

        r_ipart = scipy.zeros((pos_holder + 1, pos_holder + 1)).astype(t)
        r_ipart = R_a[: pos_holder + 1, : pos_holder + 1]

        row_i, col_i = r_ipart.shape

        #    exit1 = 'no' # 2001/04/26 changed for Python21 and future compatibility
        exit1 = 0
        L_switch = False  # added 20020416 class attribute for conservation detection 0 = none, 1 = exists

        '''If there are free variable then they are extracted and packaged, row/col vectors are formed'''

        if col - col_i > self.stoichiometric_analysis_fp_zero:
            r_fpart = scipy.zeros((pos_holder + 1, col - (pos_holder + 1))).astype(t)
            r_fpart = R_a[: pos_holder + 1, pos_holder + 1 :]

            col_vector_dependent = column_vector[pos_holder + 1 :]
            col_vector_independent = column_vector[: pos_holder + 1]
            # cons_col_vector = scipy.concatenate((col_vector_independent,col_vector_dependent),1)
            cons_col_vector = scipy.hstack(
                (col_vector_independent, col_vector_dependent)
            )
            cons_row_vector = col_vector_dependent
            # lmatrix_row_vector = scipy.concatenate((col_vector_independent,col_vector_dependent),1)
            lmatrix_row_vector = scipy.hstack(
                (col_vector_independent, col_vector_dependent)
            )
            lmatrix_col_vector = col_vector_independent
            lomatrix_row_vector = col_vector_dependent
            lomatrix_col_vector = col_vector_independent
            Nred_vector = col_vector_independent
            L_switch = True
        else:
            lomatrix_row_vector = column_vector[: pos_holder + 1]
            lomatrix_col_vector = column_vector[: pos_holder + 1]
            Nred_vector = column_vector[: pos_holder + 1]
            # print('F is empty, R = I, no conservation matrix, L is Lo')
            r_fpart = r_ipart
            #        exit1 = 'yes' # 2001/04/26 changed for Python21 and future compatibility
            exit1 = 1
            L_switch = False

        #    if exit1 == 'yes': # 2001/04/26 changed for Python21 and future compatibility
        if exit1 == 1:
            r_fpart = scipy.transpose(r_fpart)
            lmatrix = r_fpart
            # 02/10/2000 removed so that the thing returns the Lo matrix as L
            # r_fpart = 'no conservation Lo = L = I'
            # my factorization routines now swap things around for numeric stability, so that Nr might be a row/column swapped
            # this simply synchronizes Nr with its labels - brett 20050805
            Nfull = scipy.transpose(Nfull)
            row, col = Nfull.shape
            Nred = scipy.zeros((row_i, col)).astype(t)
            for x in range(0, row_i):
                Nred[x, :] = Nfull[Nred_vector[x], :]
            # return the right stuff - brett 20050805
            return (
                r_ipart,
                'no conservation',
                'no conservation',
                'no conservation',
                lmatrix,
                lomatrix_row_vector,
                lomatrix_col_vector,
                lmatrix,
                lomatrix_row_vector,
                lomatrix_col_vector,
                Nred,
                Nred_vector,
                row_vector,
                L_switch,
            )
            # return(r_ipart,'no conservation','no conservation','no conservation',lmatrix,lomatrix_row_vector,lomatrix_col_vector,lmatrix,lomatrix_row_vector,lomatrix_col_vector,scipy.transpose(Nfull),Nred_vector,row_vector,L_switch)
        else:
            r_fpart = scipy.transpose(r_fpart)
            row, col = r_fpart.shape

            id = scipy.identity(row).astype(t)
            # consmatrix = scipy.concatenate((-r_fpart,id),1).astype(t)
            consmatrix = scipy.hstack((-r_fpart, id)).astype(t)

            id = scipy.identity(col).astype(t)
            lmatrix = scipy.concatenate((id, r_fpart), 0).astype(t)

            '''This bit creates Nr. The transpose is only necessary if Nfull is already transposed in the input function'''

            Nfull = scipy.transpose(Nfull)
            row, col = Nfull.shape
            Nred = scipy.zeros((row_i, col)).astype(t)
            for x in range(0, row_i):
                Nred[x, :] = Nfull[Nred_vector[x], :]

            return (
                r_ipart,
                consmatrix,
                cons_row_vector,
                cons_col_vector,
                lmatrix,
                lmatrix_row_vector,
                lmatrix_col_vector,
                r_fpart,
                lomatrix_row_vector,
                lomatrix_col_vector,
                Nred,
                Nred_vector,
                row_vector,
                L_switch,
            )

    def SVD_Rank_Check(self, matrix=None, factor=1.0e4, resultback=0):
        """
        SVD_Rank_Check(matrix=None,factor=1.0e4,resultback=0)

        Calculates the dimensions of L/L0/K/K) by way of SVD and compares them to the Guass-Jordan results. Please note that for LARGE ill conditioned matrices the SVD can become numerically unstable when used for nullspace determinations

        Arguments:

        matrix [default=None]: the stoichiometric matrix default is self.Nmatrix
        factor [default=1.0e4]: factor used to calculate the 'zero pivot' mask = mach_eps*factor
        resultback [default=0]: return the SVD results, U, S, vh

        """
        if matrix == None:
            matrix = self.nmatrix

        nrow, ncol = matrix.shape
        TrMat = 0
        if ncol > nrow:
            # 'INFO: SVD is more accurate for tall matrices using (N)T\n'
            matrix = scipy.transpose(matrix)
            TrMat = 1

        u, s, vh = scipy.linalg.svd(matrix)
        maskF = scipy.machar.machar_double.eps * factor

        if TrMat == 0:
            print('SVD zero mask:', maskF)
            print(
                'LU factorization effective zero:',
                self.stoichiometric_analysis_lu_precision,
            )

            rank = len([el for el in (abs(s) > maskF) if el > 0.0])
            ##            null_mask = (abs(s) > maskF)
            ##            vhnull = scipy.compress(null_mask,vh,axis=0)
            ##            unull = scipy.compress(null_mask,u,axis=0)
            ##            assert vhnull.shape[0] == unull.shape[0], 'This should be true or I\'m very confused'
            ##            rank = vhnull.shape[0]

            print('\nNmatrix has ' + str(nrow) + ' rows and ' + str(ncol) + ' columns')

            print('\nSVD \"considers\" the rank to be:        ' + str(rank))
            print(
                'LU (Kmatrix) considers the rank to be: '
                + str(self.kzeromatrix.shape[0])
            )
            print(
                'LU (Lmatrix) considers the rank to be: '
                + str(self.lzeromatrix.shape[1])
            )

            print('\nComparing lzeromatrix dimensions')
            print('SVD lzeromatrix.shape =', (u.shape[0] - rank, rank))
            print('LU  lzeromatrix.shape =', self.lzeromatrix.shape)

            print('\nComparing lmatrix dimensions')
            print('SVD lmatrix.shape =', (u.shape[0], rank))
            print('LU  lmatrix.shape =', self.lmatrix.shape)

            print('\nComparing kzeromatrix dimensions')
            print('SVD kzeromatrix.shape =', (rank, vh.shape[0] - rank))
            print('LU  kzeromatrix.shape =', self.kzeromatrix.shape)

            print('\nComparing kmatrix dimensions')
            print('SVD kmatrix.shape =', (vh.shape[0], vh.shape[0] - rank))
            print('LU  kmatrix.shape =', self.kmatrix.shape)
        else:
            print('SVD zero mask:', maskF)
            print(
                'LU factorization effective zero:',
                self.stoichiometric_analysis_lu_precision,
            )

            rank = len([el for el in (abs(s) > maskF) if el > 0.0])
            ##            null_mask = (abs(s) > maskF)
            ##            vhnull = scipy.compress(null_mask,vh,axis=0)
            ##            unull = scipy.compress(null_mask,u,axis=0)
            ##            assert vhnull.shape[0] == unull.shape[0], 'This should be true or I\'m very confused'
            ##            rank = vhnull.shape[0]

            print('\nNmatrix has ' + str(nrow) + ' rows and ' + str(ncol) + ' columns')

            print('\nSVD considers the rank to be:          ' + str(rank))
            print(
                'LU (Kmatrix) considers the rank to be: '
                + str(self.kzeromatrix.shape[0])
            )
            print(
                'LU (Lmatrix) considers the rank to be: '
                + str(self.lzeromatrix.shape[1])
            )

            print('\nComparing lzeromatrix dimensions')
            print('SVD lzeromatrix.shape =', (vh.shape[0] - rank, rank))
            print('LU  lzeromatrix.shape =', self.lzeromatrix.shape)

            print('\nComparing lmatrix dimensions')
            print('SVD lmatrix.shape =', (vh.shape[0], rank))
            print('LU  lmatrix.shape =', self.lmatrix.shape)

            print('\nComparing kzeromatrix dimensions')
            print('SVD kzeromatrix.shape =', (rank, u.shape[0] - rank))
            print('LU  kzeromatrix.shape =', self.kzeromatrix.shape)

            print('\nComparing kmatrix dimensions')
            print('SVD kmatrix.shape =', (u.shape[0], u.shape[0] - rank))
            print('LU  kmatrix.shape =', self.kmatrix.shape)

        print('\nMatrix value check\n******************')

        print('\nKzeromatrix:')
        sm, bg = self.MatrixValueCompare(self.kzeromatrix)
        print('Smallest value:  ', abs(sm))
        print('Largest value:   ', abs(bg))
        print('Ratio abs(bg/sm):', abs(bg / sm))

        print('\nLzeromatrix:')
        sm, bg = self.MatrixValueCompare(self.lzeromatrix)
        print('Smallest value:  ', abs(sm))
        print('Largest value:   ', abs(bg))
        print('Ratio abs(bg/sm):', abs(bg / sm))

        print(
            '\nPlease note: I\'ve found that for larger models the rank calculated by SVD in this test can become unstable and give incorrect results.'
        )
        print(' ')

        if resultback:
            return u, s, vh


if __psyco_active__:
    psyco.bind(MathArrayFunc)
    psyco.bind(Stoich)
